# system modules
import datetime
import functools
import inspect
import itertools
import os
import re
import sqlite3
from abc import ABC, abstractmethod, abstractproperty

from sensemapi.reprobject import ReprObject
from sensemapi.utils import *
from sensemapi.xdg import *

# external modules


class AuthentiCache(ABC, ReprObject):
    """
    Abstract base class for authentication caches

    Args:
        directory (str, optional): the cache directory
        name (str, optional): the cache name
    """

    def __init__(self, directory=None, name=None):
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)

    @simplegetter
    def directory(self):
        return XDGPackageDirectory("XDG_CACHE_HOME").path

    @simplesetter(directory)
    def directory(self, new):
        return Directory(new).path

    @simplegetter
    def name(self):
        return "authentication"

    @simplesetter(name)
    def name(self, new):  # pragma: no cover
        return str(new)

    @abstractproperty
    def suffix(self):
        pass

    @abstractmethod
    def get(
        self,
        conditions,
        all_inner=True,
        all_outer=True,
        match_inner=True,
        match_outer=True,
    ):
        """
        Given conditions, retrieve matching cache values

        Args:
            conditions (sequence of dict, optional):
                dictionaries mapping column names to expected values.
            all_inner (sequence of bool, optional): whether all conditions
                specified in a single condition in ``conditions`` should be
                matched together
            all_outer (sequence of bool, optional): whether all ``conditions``
                should be matched together
            match_inner (sequence of bool, optional): whether all conditions
                specified in a single condition in ``conditions`` should be
                matched (or negated)
            match_outer (sequence of bool, optional): whether all
                ``conditions`` should be matched (or negated)

        Returns:
            sequence : the matching entries
        """
        pass

    @abstractmethod
    def delete(self, condition=()):
        """
        Delete cache entr(y/ies) matching condition. If called without
        arguments, deletes the whole cache.
        """
        pass

    @abstractmethod
    def sync(self, cache):
        """
        Given cache values, update the cache where necessary and return
        up-to-date values.

        Args:
            cache (dict) : cache data with any of keys email, username, api,
                token, token_time, refresh_token, refresh_token_time. Keys that
                are not provided are ignored.

        Returns
            dict : the updated cache data
        """
        pass


class SQLiteAuthentiCache(AuthentiCache):
    """
    Authentication cache using :mod:`sqlite3` as storage.
    """

    @staticmethod
    def sanitize(string, brutally=False):
        return re.sub(r"\W" if brutally else "['\";]", "", string)

    @property
    def suffix(self):
        return "sqlite"

    @property
    def connection(self):
        with Directory(self.directory) as d:
            conn = sqlite3.connect(
                os.path.join(d, "{}.{}".format(self.name, self.suffix)),
                detect_types=sqlite3.PARSE_DECLTYPES,
            )
            conn.row_factory = sqlite3.Row
            return conn

    @property
    def table(self):
        return "authentication"

    @property
    def table_exists(self):
        with self.connection as conn:
            return any(
                self.table in row
                for row in conn.execute(
                    "SELECT name FROM sqlite_master WHERE type='table'"
                )
            )

    @property
    def tableheader(self):
        self.create_table()
        with self.connection as conn:
            cursor = conn.execute(
                "SELECT * from {cache.table}".format(cache=self)
            )
            return tuple(map(lambda x: x[0], cursor.description))

    @property
    def primary_keys(self):
        return ("email", "username", "api")

    def create_table(self):
        if not self.table_exists:
            with self.connection as conn:
                sql = (
                    "CREATE TABLE {cache.table} "
                    "(email, username, password, api, "
                    "token, token_time timestamp, "
                    "refresh_token, refresh_token_time timestamp, "
                    "PRIMARY KEY ({primary_keys}) "
                    ")"
                ).format(cache=self, primary_keys=",".join(self.primary_keys))
                logger.debug("SQL: {}".format(sql))
                conn.execute(sql)
                return True
        return False

    def sql_insert(self, row):
        """
        Generate SQL expression to insert

        Args:
            rows (dict): the row to insert with keys corresponding to the
                columns

        Returns:
            str, sequence : the SQL expression to use after
                ``INSERT INTO table...`` and the corresponding parameters
        """
        return (
            "({columns}) VALUES ({values})".format(
                columns=",".join(row.keys()), values=",".join("?" for c in row)
            ),
            tuple(row.values()),
        )

    def sql_update(self, old_row, new_row):
        """
        Generate SQL expression to update
        """
        assert old_row and new_row, "both old_row and new_row mustn't be empty"
        sql_cond, sql_cond_params = self.sql_condition(old_row)
        return (
            "SET {assignments} WHERE {condition}".format(
                assignments=",".join(map("{} = ?".format, new_row.keys())),
                condition=sql_cond,
            ),
            tuple(new_row.values()) + sql_cond_params,
        )

    def sql_condition(
        self,
        conditions=tuple(),
        all_inner=True,
        all_outer=True,
        match_inner=True,
        match_outer=True,
    ):
        """
        Generate an SQL conditional expression according to exact conditions.
        All arguments are :func:`itertools.cycle` d accordingly.

        Args:
            conditions (sequence of dict, optional):
                dictionaries mapping column names to expected values. If this
                sequence is empty, a condition that is always true is returned.
            all_inner (sequence of bool, optional): whether all conditions
                specified in a single condition in ``conditions`` should be
                matched together
            all_outer (sequence of bool, optional): whether all ``conditions``
                should be matched together
            match_inner (sequence of bool, optional): whether all conditions
                specified in a single condition in ``conditions`` should be
                matched (or negated)
            match_outer (sequence of bool, optional): whether all
                ``conditions`` should be matched (or negated)

        Returns:
            str, sequence : an SQL expression which can be used with ``WHERE``
                and the corresponding parameters
        """
        args = locals()
        # put the conditions into a dict
        conds = {
            pref: args.get(pref)
            for pref in (
                "all_inner",
                "all_outer",
                "match_inner",
                "match_outer",
            )
        }
        conds = {  # properly cycle all settings
            pref: itertools.cycle(val if hasattr(val, "__iter__") else (val,))
            for pref, val in conds.items()
        }
        conds["values"] = (
            (conditions,) if hasattr(conditions, "items") else conditions
        )
        sql_conditions, sql_params = [], []
        for cond in listdict_to_dictlist(conds):
            sanitized_values = {
                self.sanitize(k, brutally=True): self.sanitize(
                    v, brutally=False
                )
                if isinstance(v, str)
                else v
                for k, v in cond["values"].items()
                if k in self.tableheader
            }
            inner_sql_cond = (" AND " if cond["all_inner"] else " OR ").join(
                map(
                    "{} {} ?".format,
                    sanitized_values.keys(),
                    itertools.cycle(
                        ("IS" if cond["match_inner"] else "IS NOT",)
                    ),
                )
            ) or "1=1"
            sql_conditions.append(
                "{}({})".format(
                    "" if cond["match_outer"] else "NOT ", inner_sql_cond
                )
            )
            sql_params.extend(tuple(sanitized_values.values()))
        sql_condition = (
            functools.reduce(
                lambda x, y: (
                    " AND " if next(conds["all_outer"]) else " OR "
                ).join((x, y)),
                sql_conditions,
            )
            if sql_conditions
            else "1=1"
        )
        return sql_condition, tuple(sql_params)

    def matching_rows(self, condition):
        """
        Perform greedy matching, then less greedy matching to find matching
        rows.

        Args:
            condition (dict): dictionary of column names and values

        Returns:
            sequence : the matching row(s)
        """

        def two_primary_keys_and_api(seq):
            return (
                len(set(self.primary_keys).intersection(seq)) >= 2
                and "api" in seq
            )

        for i in range(len(condition) - 1):
            for comb in filter(
                two_primary_keys_and_api,
                itertools.combinations(condition, len(condition) - i),
            ):
                logger.debug("Checking combination {}".format(comb))
                cond = {k: condition.get(k) for k in comb}
                matching = self.get(cond)
                if matching:
                    logger.debug(
                        "Matching entr(y/ies) for "
                        "combination {}:\n{}".format(comb, matching)
                    )
                    return matching
                else:
                    logger.debug("No entries for combination {}".format(comb))
        return tuple()

    def get(self, *args, **kwargs):
        """
        Retrieve cache according to conditions. For arguments, see
        :meth:`sql_condition`.
        """
        sql_cond, sql_params = self.sql_condition(*args, **kwargs)
        with self.connection as conn:
            sql = "SELECT * FROM {cache.table} {condition}".format(
                condition=("WHERE " + sql_cond) if sql_cond else sql_cond,
                cache=self,
            )
            logger.debug("SQL:\n{}\nparameters: {}".format(sql, sql_params))
            return tuple(map(dict, conn.execute(sql, sql_params)))

    def delete(self, condition=()):
        sql_cond, sql_params = self.sql_condition(condition)
        with self.connection as conn:
            sql = "DELETE FROM {cache.table} {condition}".format(
                condition=("WHERE " + sql_cond) if sql_cond else sql_cond,
                cache=self,
            )
            logger.debug("SQL:\n{}\nparameters: {}".format(sql, sql_params))
            return conn.execute(sql, sql_params)

    def sync(self, cache):
        logger.debug("Trying to determine entry matching {}".format(cache))
        primary_key_equal = {
            k: v
            for k, v in cache.items()
            if k in self.primary_keys and v is not None
        }
        matching_rows = self.matching_rows(primary_key_equal)
        logger.debug("Matching cache entr(y/ies):\n{}".format(matching_rows))
        ancient = datetime.datetime.min
        if len(matching_rows) > 1:
            logger.warning(
                "There are duplicate rows:\n{}\n"
                "Using the first one.".format(matching_rows)
            )
        try:
            matching_row = matching_rows[0]
        except IndexError:
            matching_row = matching_rows
        if matching_row:
            to_save = cache.copy()
            our_to = matching_row.get("token")
            their_to = cache.get("token")
            our_reto = matching_row.get("refresh_token")
            their_reto = cache.get("refresh_token")
            our_toti = matching_row.get("token_time", ancient) or ancient
            our_retoti = (
                matching_row.get("refresh_token_time", ancient) or ancient
            )
            their_toti = cache.get("token_time", ancient) or ancient
            their_retoti = cache.get("refresh_token_time", ancient) or ancient
            their_to_younger = their_toti > our_toti
            logger.debug(
                "Given token '{}' from '{}' is {}newer than "
                "ours from ('{}').".format(
                    their_to,
                    their_toti,
                    "" if their_to_younger else "not ",
                    our_toti,
                )
            )
            if their_to_younger:
                logger.debug("Using newer given token '{}'".format(their_to))
                to_save["token"] = their_to
                to_save["token_time"] = their_toti
            else:
                logger.debug("Ignoring given token '{}'".format(their_to))
                to_save.pop("token", None)
                to_save.pop("token_time", None)
            their_reto_younger = their_retoti > our_retoti
            logger.debug(
                "Given refresh token '{}' from '{}' is {}newer than "
                "ours from ('{}').".format(
                    their_reto,
                    their_retoti,
                    "" if their_reto_younger else "not ",
                    our_retoti,
                )
            )
            if their_reto_younger:
                logger.debug(
                    "Using newer given refresh token '{}'".format(their_reto)
                )
                to_save["refresh_token"] = their_reto
                to_save["refresh_token_time"] = their_retoti
            else:
                logger.debug(
                    "Ignoring given refresh token '{}'".format(their_reto)
                )
                to_save.pop("refresh_token", None)
                to_save.pop("refresh_token_time", None)
            if to_save.items() <= matching_row.items():
                logger.debug(
                    "No need to update\n{}\nwith\n{},"
                    "\nno new information".format(matching_row, to_save)
                )
                return matching_row
            logger.debug(
                "Will update\n{}\nwith\n{}".format(matching_row, to_save)
            )
            sql_update, sql_params = self.sql_update(matching_row, to_save)
            sql = "UPDATE {cache.table} {update}".format(
                cache=self, update=sql_update
            )
            logger.debug("SQL:\n{}\nparameters:{}".format(sql, sql_params))
            with self.connection as conn:
                conn.execute(sql, sql_params)
            updated_row = matching_row.copy()
            updated_row.update(to_save)
            return updated_row
        else:
            logger.debug("Will create entry {}".format(cache))
            sql_insert_data, sql_params = self.sql_insert(cache)
            sql_insert = "INSERT INTO {cache.table} {data}".format(
                cache=self, data=sql_insert_data
            )
            logger.debug(
                "SQL:\n{}\nparameters: {}".format(sql_insert, sql_params)
            )
            with self.connection as conn:
                conn.execute(sql_insert, sql_params)
            return cache



# system modules
import functools
import inspect
import io
import itertools
import json
import logging
import math
import re
import time
from contextlib import contextmanager
from posixpath import join as urljoin

# external modules
import requests

# internal modules
from sensemapi import compat, paths
from sensemapi.errors import *
from sensemapi.reprobject import ReprObject
from sensemapi.senseBox import senseBox
from sensemapi.senseBoxSensorData import senseBoxSensorData
from sensemapi.utils import *

logger = logging.getLogger(__name__)


class SenseMapClient(ReprObject):
    """
    Client to interface the `OpenSenseMap API <https://api.opensensemap.org>`_

    Args:
        api (str, optional): the api server to use. You may use
            :any:`OPENSENSEMAP_API_LIVE` (default) or
            :any:`OPENSENSEMAP_API_TEST` for testing purposes.
        connection_timeout (float, optional): default connection timeout
            for requests in seconds. Default is no timeout.
        request_cache (cachecontrol.adapter.CacheControl, optional):
            request cache
        response_timeout (float, optional): default response timeout
            for requests in seconds. Default is no timeout.
    """

    def __init__(
        self,
        api=paths.OPENSENSEMAP_API_LIVE,
        connection_timeout=None,
        response_timeout=None,
        request_cache=None,
    ):
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)

    RETRY_AFTER_HEADER_REGEX = re.compile("retry.*after$", re.IGNORECASE)
    """
    Regex used to determine the header field containing the time to wait until
    issuing the next request
    """

    def modifies_session(decorated_function):
        """
        Decorator for methods that require closing and deleting the
        :any:`SenseMapClient.session` because they modify it.
        """

        @functools.wraps(decorated_function)
        def wrapper(self, *args, **kwargs):
            if hasattr(self, "_session"):
                logger.debug(
                    "Closing and deleting existing session "
                    "because {} was called".format(decorated_function)
                )
                self._session.close()
                del self._session
            return decorated_function(self, *args, **kwargs)

        return wrapper

    @simplegetter
    def api(self):
        return None

    @api.setter
    @modifies_session
    def api(self, new):
        self._api = new

    @simplegetter
    def connection_timeout(self):
        return None

    @simplesetter(connection_timeout)
    def connection_timeout(self, new):
        return new if new is None else float(new)

    @simplegetter
    def response_timeout(self):
        return None

    @simplesetter(response_timeout)
    def response_timeout(self, new):
        return new if new is None else float(new)

    @simplegetter
    @needs_cachecontrol
    def cache_heuristic(self):
        from sensemapi.heuristic import CacheHeuristic

        return CacheHeuristic()

    @cache_heuristic.setter
    @modifies_session
    def cache_heuristic(self, new):
        self._cache_heuristic = new

    @simplegetter
    def session(self):
        session = requests.session()
        if self.request_cache:
            try:
                from cachecontrol.adapter import CacheControlAdapter

                adapter = CacheControlAdapter(
                    cache=self.application.request_cache
                    if self.application
                    else self.request_cache,
                    heuristic=self.cache_heuristic,
                    cacheable_methods=("GET",),
                )
                if self.api:
                    # we will only cache GET requests to the /boxes path
                    session.mount(urljoin(self.api, paths.BOXES), adapter)
                else:
                    logger.warning(
                        "No API is set. "
                        "Mounting cache adapter for whole http(s):// schema"
                    )
                    for schema in ("http://", "https://"):
                        session.mount(schema, adapter)
            except ImportError:
                logger.warning(
                    "CacheControl is not available. Cannot use cache."
                )
        return session

    @session.setter
    @modifies_session
    def session(self, new):
        self._session = new

    @simplegetter
    def request_cache(self):
        return None

    @request_cache.setter
    @modifies_session
    def request_cache(self, new):
        self._request_cache = new

    @property
    @contextmanager
    def no_cache(self):
        """
        Context manager to disable caching temporarily by setting the default
        'Cache-Control' header to 'no-cache'.
        """
        old_cache_control = self.session.headers.get("Cache-Control")
        logger.debug(
            "Temporarily setting default 'Cache-Control' header to 'no-cache'"
        )
        self.session.headers.update({"Cache-Control": "no-cache"})
        yield
        if old_cache_control:
            logger.debug(
                "Resetting default 'Cache-Control' header "
                "from '{}' to {}".format(
                    self.session.headers.get("Cache-Control"),
                    old_cache_control,
                )
            )
        else:
            logger.debug("Removing default 'Cache-Control' header")
        self.session.headers.pop("Cache-Control")

    def retry_after_time_if_too_many_requests(decorated_function):
        """
        Decorator to call a decorated method, catching an
        :any:`OpenSenseMapAPITooManyRequestsError`, trying to determine how
        long to wait from the error message, waiting that time and then
        retries.
        """

        @functools.wraps(decorated_function)
        def wrapper(self, *args, **kwargs):
            try:
                return decorated_function(self, *args, **kwargs)
            except OpenSenseMapAPITooManyRequestsError as e:
                logger.debug(e)
                m = re.search(
                    pattern=r"(?:(?P<seconds>\d+\.\d+)\s*s)|"
                    r"(?:(?P<milliseconds>\d+\.\d+)\s*ms)",
                    string=str(e),
                )
                try:
                    d = m.groupdict()
                    retry_after_seconds = 0
                    seconds = d.get("seconds")
                    milliseconds = d.get("milliseconds")
                    if seconds:
                        retry_after_seconds += float(seconds)
                    if milliseconds:
                        retry_after_seconds += float(milliseconds) / 1000
                except (AttributeError, ValueError, KeyError, IndexError):
                    raise OpenSenseMapAPITooManyRequestsError(
                        "Could not determine " "time to wait until next retry."
                    )
                logger.debug(
                    "Waiting {} seconds until retry...".format(
                        retry_after_seconds
                    )
                )
                time.sleep(retry_after_seconds)
            logger.debug(
                "Now trying again to call {}".format(decorated_function)
            )
            return decorated_function(self, *args, **kwargs)

        return wrapper

    @retry_after_time_if_too_many_requests
    def request(self, method, *args, **kwargs):
        """
        Wrapper around corresponding methods of :mod:`requests`, raising
        specific exceptions depending of the response.

        Args:
            method (str): the method to use. Needs to be a method of
                :any:`requests`.
            args, kwargs : arguments passed to the method

        Returns:
            requests.models.Response : the request response

        Raises:
            OpenSenseMapAPITooManyRequestsError : if the client issued too many
                requests
        """
        request_kwargs = {
            "timeout": (self.connection_timeout, self.response_timeout)
        }
        request_kwargs.update(kwargs)
        response = self.session.request(method, *args, **request_kwargs)
        logger.debug(
            "API responded [status code {}]:\n"
            "Headers:\n{}\nPayload:\n{}".format(
                response.status_code, response.headers, response.text
            )
        )
        # too many requests
        if response.status_code == 429:
            headers = response.headers.copy()
            retry_after_header = next(
                filter(self.RETRY_AFTER_HEADER_REGEX.search, headers.keys()),
                None,
            )
            retry_after = headers.get(retry_after_header)
            raise OpenSenseMapAPITooManyRequestsError(
                "retry after {}".format(retry_after) if retry_after else ""
            )
        try:
            logger.debug(
                "API responded JSON [status code {}]:\n{}".format(
                    response.status_code, response.json()
                )
            )
        except compat.JSONDecodeError:
            pass
        return response

    def _get_box(self, id, format=None):
        """
        Issue the request to retreive a single senseBox

        Args:
            id (str) : the senseBox id to retreive
            format (str, optional): one of ``"json"`` and ``"geojson"``

        Returns:
            dict : the API response
        """
        response = self.request(
            "get",
            urljoin(self.api, paths.BOXES, id),
            params={"format": format} if format else {},
        )
        response_json = response.json()
        if response.status_code == 200:
            return response_json
        else:
            message = response_json.get("message")
            raise OpenSenseMapAPIError(
                "Could not retreive with id '{}'{}".format(
                    id, ": {}".format(message) if message else ""
                )
            )

    def get_box(self, id):
        """
        Retreive one :class:`senseBox`

        Args:
            id (str) : the :class:`senseBox` id to retreive

        Returns:
            senseBox : the retreived senseBox
        """
        box = senseBox.from_api_json(self._get_box(id=id, format="json"))
        box.client = self
        return box

    def post_measurement(
        self,
        box_id,
        sensor_id,
        value,
        time=None,
        lat=None,
        lon=None,
        height=None,
    ):
        """
        Issue a request to upload a new measurement

        Args:
            box_id (str) : the senseBox id
            sensor_id (str) : the sensor's id
            value (float) : the current measurement value
            time (datetime.datetime, optional) : the time of the measurement
            lat, lon, height (float,optional) : the current position

        Returns:
            True : on success
        """
        assert box_id is not None, "box_id must  be defined"
        assert sensor_id is not None, "sensor_id must  be defined"
        d = {}
        d["value"] = float(value)
        if time:
            d["createdAt"] = date2str(time)
        try:
            d["location"] = location_dict(lat, lon, height)
        except ValueError:
            pass
        logger.debug("Sending Request with JSON:\n{}".format(pretty_json(d)))
        response = self.request(
            "post", urljoin(self.api, paths.BOXES, box_id, sensor_id), json=d
        )
        try:
            response_json = response.json()
        except compat.JSONDecodeError:  # pragma: no cover
            raise OpenSenseMapAPIError(
                "Posting measurement didn't work: {}".format(response.text)
            )
        if hasattr(response_json, "get"):  # is a dict
            message = response_json.get("message")
            raise OpenSenseMapAPIError(
                "Posting measurement didn't work{}".format(
                    ": " + message or ""
                )
            )
        else:  # no dict
            if re.search(r"measurement\s+saved\s+in\s+box", response_json):
                return True

    @classmethod
    @needs_pandas
    def dataframe_to_csv_for_upload(self, df, discard_incomplete=False):
        """
        Convert a dataframe to csv for the upload

        Args:
            df (pandas.DataFrame): a dataframe with the at least the
                columns ``sensor_id`` (string), ``value`` (float) and
                optionally ``time`` (datetime), ``lat`` (float) **and** ``lon``
                (float) and ``height`` (float).
            discard_incomplete (bool, optional): use as much data as possible.
                If ``True``, drops parts of incomplete datasets for the API to
                accept the data. If ``False`` (default), raises
                :any:`ValueError` if the dataset is incomplete.

        Returns:
            str : the csv string

        Raises
            ValueError : if ``discard_incomplete=False`` and any row is
                incomplete
        """
        data = df.copy()
        # convert time to utc
        if "time" in data:
            try:  # timezone-unaware
                data["time"] = data["time"].dt.tz_localize("UTC")
            except TypeError:  # already timezone-aware
                data["time"] = data["time"].dt.tz_convert("UTC")
        duplicated = list(
            df["sensor_id"][df["sensor_id"].duplicated(keep=False)].unique()
        )
        if duplicated:
            raise ValueError(
                "The sensor id{s1} {ids} occur{s2} more than once. "
                "Posting 'multiple' new measurements does not mean "
                "'upload of timeseries'".format(
                    s1="s" if len(duplicated) > 1 else "",
                    s2="" if len(duplicated) > 1 else "s",
                    ids=",".join(duplicated),
                )
            )
        groups = list(
            itertools.accumulate(
                (
                    ("sensor_id", "value"),
                    ("time",),
                    ("lon", "lat"),
                    ("height",),
                )
            )
        )
        csv_string = ""
        for group, smaller_group in zip(
            reversed(groups), list(reversed(groups))[1:] + [[]]
        ):
            logger.debug("current group: {}".format(group))
            logger.debug("smaller group: {}".format(smaller_group))
            try:
                part = data[list(group)]
            except KeyError as e:
                logger.debug("data is missing columns {}".format(e))
                continue
            subset = set(group) - set(smaller_group)
            part_smaller_kept = part.dropna(
                axis="index", how="all", subset=subset
            )
            part_smaller_dropped = part.drop(part_smaller_kept.index)
            if not part_smaller_dropped.empty:
                logger.debug(
                    "These {} rows are incomplete "
                    "and obviously for smaller group {}:\n{}".format(
                        len(part_smaller_dropped.index),
                        subset,
                        part_smaller_dropped,
                    )
                )
            part_kept = part_smaller_kept.dropna(axis="index", how="any")
            part_dropped = part_smaller_kept.drop(part_kept.index)
            if not part_dropped.empty:
                logger.warning(
                    "These {} rows are incomplete, "
                    "this {} data will be discarded:\n{}".format(
                        len(part_dropped.index), subset, part_dropped
                    )
                )
                if not discard_incomplete:
                    raise ValueError(
                        "In these rows, "
                        "none of columns {} must be NaN:\n{}".format(
                            smaller_group, part_dropped
                        )
                    )
            logger.debug(
                "Keeping these {} rows for group {}:\n{}".format(
                    len(part_kept.index), group, part_kept
                )
            )
            if part_kept.empty:
                continue
            csv_string += part_kept.to_csv(
                header=False,
                index=False,
                sep=",",
                date_format=OPENSENSEMAP_DATETIME_FORMAT_UTC,
            )
            data.drop(part_kept.index, inplace=True)
        if not data.empty:
            logger.warning(
                "Skipping these {} rows of data "
                "due to missing information:\n{}".format(len(data.index), data)
            )
        return csv_string

    def post_measurements(
        self, box_id, measurements, discard_incomplete=False
    ):
        """
        Post multiple measurements to sensors in a :class:`senseBox`

        Args:
            box_id (str): the senseBox id
            measurements (pandas.DataFrame): a dataframe with the at least the
                columns ``sensor_id`` (string), ``value`` (float) and
                optionally ``time`` (datetime), ``lat`` (float) **and** ``lon``
                (float) and ``height`` (float).
            discard_incomplete (bool, optional): use as much data as possible.
                If ``True``, drops parts of incomplete datasets for the API to
                accept the data. If ``False`` (default), raises
                :any:`ValueError` if the dataset is incomplete.

        Returns:
            bool : whether the upload was successful
        """
        csv_string = self.dataframe_to_csv_for_upload(
            measurements, discard_incomplete=discard_incomplete
        )
        logger.debug(
            "Attempting to upload CSV data to senseBox '{}':\n{}".format(
                box_id, csv_string
            )
        )
        response = self.request(
            "post",
            url=urljoin(self.api, paths.BOXES, box_id, "data"),
            headers={"content-type": "text/csv"},
            data=csv_string,
        )
        if response.status_code == requests.codes.CREATED:
            return True
        try:
            response_json = response.json()
        except compat.JSONDecodeError:  # pragma: no cover
            raise OpenSenseMapAPIError(
                "Posting measurement didn't work: {}".format(response.text)
            )
        message = response_json.get("message")
        raise OpenSenseMapAPIError(
            "Posting measurement didn't work{}".format(": " + message or "")
        )

    def get_measurements(
        self,
        box_id,
        sensor_id,
        from_date=None,
        to_date=None,
        format=None,
        download=None,
        outliers=None,
        outlier_window=None,
        delimiter=None,
    ):
        """
        Retrieve the 10000 latest measurements for a sensor

        Args:
            box_id (str): the senseBox id
            sensor_id (str): the sensor id
            from_date (datetime.datetime, optional): beginning date of
                measurement data (default: 48 hours ago from now)
            to_date (datetime.datetime, optional): end date of measurement data
                (default: now)
            format (str, optional): either ``"json"`` (default) or ``"csv"``
            outliers (bool, optional): add outlier marker `isOutlier` to data?
            outlier_window (int, optional):
                outlier window size (1-50, default: 15)
            delimiter (str, optional): either ``"comma"`` (default) or
                ``"semicolon"``

        Returns:
            senseBoxSensorData : the retrieved data
        """
        assert box_id is not None, "box_id must  be defined"
        assert sensor_id is not None, "sensor_id must  be defined"
        d = {}
        if from_date is not None:
            d["from-date"] = date2str(from_date)
        if to_date is not None:
            d["to-date"] = date2str(to_date)
        if format is not None:
            d["format"] = str(format)
        if outliers is not None:
            d["outliers"] = bool(outliers)
        if outlier_window is not None:
            d["outlier-window"] = int(outlier_window)
        logger.debug(
            "Sending GET request with parameters:\n{}".format(pretty_json(d))
        )
        response = self.request(
            "get",
            urljoin(self.api, paths.BOXES, box_id, "data", sensor_id),
            params=d,
        )
        try:
            response_json = response.json()
            raise NotImplementedError(
                "Parsing JSON-formatted measurements "
                "is not yet implemented. Use format='csv'"
            )
        except compat.JSONDecodeError:
            return senseBoxSensorData.from_csv(io.StringIO(response.text))

    def __del__(self):
        """
        Class destructor. Calls :meth:`requests.Session.close` on
        :any:`SenseMapClient.session`.
        """
        self.session.close()



# system modules
import argparse

# internal modules
import sensemapi

# external modules

parser = argparse.ArgumentParser(
    prog="python3 -m sensemapi",
    description="Command-line interface to the OpenSenseMap API",
    epilog="This CLI is not yet implemented.",
)
args = parser.parse_args()
parser.print_help()



# external modules
import requests
from requests.compat import json as requestsjson

try:  # requestsjson
    JSONDecodeError = requestsjson.JSONDecodeError
except AttributeError:  # normal json
    try:  # newer Python
        JSONDecodeError = requestsjson.errors.JSONDecodeError
    except AttributeError:  # older Python (3.4)
        JSONDecodeError = ValueError



class SenseMapiError(Exception):
    """
    Base Exception class for package errors
    """


class NoCredentialsError(SenseMapiError):
    """
    Base SenseMapiError class raised whenever credentials are missing
    """


class NoUserError(NoCredentialsError):
    """
    SenseMapiError raised when no user/email is available
    """


class NoEmailError(NoUserError):
    """
    SenseMapiError raised when no email is available
    """


class NoUsernameError(NoUserError):
    """
    SenseMapiError raised when no user is available
    """


class NoPasswordError(NoCredentialsError):
    """
    SenseMapiError raised when no password is available
    """


class OpenSenseMapAPIError(SenseMapiError):
    """
    Base SenseMapiError class raised whenever something is strange
    """

    pass


class OpenSenseMapAPIResponseError(SenseMapiError):
    """
    SenseMapiError raised when the API responded something strange
    """

    pass


class OpenSenseMapAPIAuthenticationError(OpenSenseMapAPIError):
    """
    SenseMapiError raised when the authentication failed
    """

    pass


class OpenSenseMapAPIInvalidCredentialsError(
    OpenSenseMapAPIAuthenticationError
):
    """
    SenseMapiError raised when the login failed due to invalid credentials
    """

    pass


class OpenSenseMapAPIOutdatedTokenError(OpenSenseMapAPIAuthenticationError):
    """
    SenseMapiError raised when a token is outdated
    """

    pass


class OpenSenseMapAPITooManyRequestsError(OpenSenseMapAPIError):
    """
    SenseMapiError raised when a client issues too many requests
    """

    pass


class OpenSenseMapAPIPermissionError(OpenSenseMapAPIAuthenticationError):
    """
    SenseMapiError raised when the account does not have certain permissions
    """

    pass


class NoClientError(SenseMapiError):
    pass


class ConfirmationError(SenseMapiError):
    pass


class NoPandasError(ImportError):
    """
    SenseMapiError raised when no :mod:`pandas` was found
    """

    def __init__(self, message=""):
        super().__init__(message or "pandas is not installed")


class NoCacheControl(ImportError):
    """
    SenseMapiError raised when no :mod:`cachecontrol` was found
    """

    def __init__(self, message=""):
        super().__init__(message or "cachecontrol is not installed")



# system modules
import contextlib
import logging
import os
from os.path import abspath, exists, expanduser, join

from sensemapi import __name__ as packagename
from sensemapi.reprobject import ReprObject
from sensemapi.utils import *

# external modules

logger = logging.getLogger(__name__)

XDG_HOME_DEFAULTS = {
    "XDG_DATA_HOME": expanduser(join("~", ".local", ".share")),
    "XDG_CONFIG_HOME": expanduser(join("~", ".config")),
    "XDG_CACHE_HOME": expanduser(join("~", ".cache")),
}
XDG_DIRS_DEFAULTS = {
    "XDG_DATA_DIRS": ":".join(
        (
            expanduser(join(join(os.sep, "usr"), "local", "share")),
            expanduser(join(join(os.sep, "usr"), "share")),
        )
    ),
    "XDG_CONFIG_DIRS": expanduser(join(join(os.sep, "etc"), "xdg")),
}
XDG_DEFAULTS = XDG_HOME_DEFAULTS.copy()
XDG_DEFAULTS.update(XDG_DIRS_DEFAULTS)


class XDGDirectoryVariable(ReprObject):
    def __init__(self, name, defaults=XDG_DEFAULTS):
        self.defaults = defaults
        self.name = name

    @simplegetter
    def name(self):  # pragma: no cover
        return ""

    @simplesetter(name)
    def name(self, new):
        new = str(new)
        if new not in self.defaults:
            raise ValueError(
                "Invalid name '{}'. Use one of {}.".format(
                    new, ", ".join(map("'{}'".format, self.defaults.keys()))
                )
            )
        return new

    @property
    def value(self):
        return os.environ.get(
            self.name, XDG_DEFAULTS.get(self.name)
        ) or XDG_DEFAULTS.get(self.name)

    def __str__(self):
        return str(self.value)


class Directory(ReprObject):
    def __init__(self, path):
        self.path = path

    def __enter__(self):
        if not exists(self.path):
            logger.info("Create nonexistent directory '{}'".format(self.path))
            os.makedirs(self.path)
        return self.path

    def __exit__(self, *args, **kwargs):
        pass

    def __str__(self):
        return str(self.path)


class XDGDirectory(XDGDirectoryVariable, Directory):
    def __init__(self, name):
        XDGDirectoryVariable.__init__(
            self, name=name, defaults=XDG_HOME_DEFAULTS
        )

    @property
    def path(self):
        return expanduser(self.value)


class XDGPackageDirectory(XDGDirectory):
    def __init__(self, name, packagename=__name__.split(".")[0]):
        XDGDirectory.__init__(self, name)
        self.packagename = packagename

    @property
    def path(self):
        return join(expanduser(self.value), self.packagename)


class XDGDirectories(XDGDirectoryVariable):
    def __init__(self, name):
        XDGDirectoryVariable.__init__(
            self, name=name, defaults=XDG_DIRS_DEFAULTS
        )

    @property
    def paths(self):
        return self.value.split(":")

    @property
    def existing_paths(self):
        return filter(
            exists,
            unique(filter(lambda p: abspath(expanduser(p)), self.paths)),
        )

    def __iter__(self):
        return self.existing_paths



OPENSENSEMAP_API_LIVE = "https://api.opensensemap.org/"
"""
The live OpenSeneMap API server
"""

OPENSENSEMAP_API_TEST = "https://api.testing.opensensemap.org/"
"""
The test OpenSeneMap API server
"""

SIGN_IN = "users/sign-in"
SIGN_OUT = "users/sign-out"
REFRESH = "users/refresh-auth"
USER_DETAILS = "users/me"
BOXES = "boxes"



# system modules
import collections
import configparser
import copy
import functools
import itertools
import logging
import os
import re

from sensemapi.account import SenseMapAccount, SenseMapAccountCollection
from sensemapi.cache import SQLiteAuthentiCache
from sensemapi.reprobject import ReprObject
from sensemapi.utils import *
from sensemapi.xdg import *

# external modules


logger = logging.getLogger(__name__)


class SenseMapApplication(ReprObject):
    """
    Class for applications using :mod:`sensemapi`
    """

    def __init__(self, name=None, version=None, accounts=None):
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)
        self.read_config()

    @simplegetter
    def name(self):
        return self.__class__.__name__

    @property
    def sanitized_name(self):
        return re.sub("--+", "-", re.sub(r"\W", "-", self.name)).lower()

    @simplesetter(name)
    def name(self, new):
        return str(new or self.__class__.__name__)

    @simplegetter
    def version(self):
        return None

    @simplesetter(version)
    def version(self, new):
        return new

    @simplegetter
    def config(self):
        return configparser.ConfigParser()

    @simplesetter(config)
    def config(self, new):
        return new

    @property
    def app_config(self):
        if "application" not in self.config:
            self.config.add_section("application")
        return self.config["application"]

    @simplegetter
    def accounts(self):
        accounts = SenseMapAccountCollection()
        accounts.application = self
        return accounts

    @simplesetter(accounts)
    def accounts(self, new):
        new.application = self
        return new

    @property
    def current_account_index(self):
        try:
            index = self._current_account_index
        except AttributeError:
            index = None
        if self.accounts:
            if index not in range(len(self.accounts)):
                index = 0
                self._current_account_index = index
            else:
                # use the setter to save the changes
                self.current_account_index = index
        else:
            index = None
            self._current_account_index = index
        return self._current_account_index

    @current_account_index.setter
    def current_account_index(self, new):
        try:
            self.accounts[new]
        except (IndexError, ValueError, KeyError, TypeError) as e:
            logger.warning(
                "Cannot use '{}' as current account index: {}".format(new, e)
            )
            new = 0 if self.accounts else None
        self._current_account_index = new
        old_current_account_index = self.app_config.getint(
            "current_account_index"
        )
        if new is not None:
            self.app_config["current_account_index"] = str(new)
        else:
            self.app_config.pop("current_account_index", None)
        if old_current_account_index != new:
            self.write_config()

    @property
    def current_account(self):
        if len(self.accounts):
            if self.current_account_index is not None:
                try:
                    return self.accounts[self.current_account_index]
                except (IndexError, ValueError, KeyError, TypeError):
                    if hasattr(self, "_current_account_index"):
                        del self._current_account_index
                return self.accounts[self.current_account_index]
        else:
            return None

    @current_account.setter
    def current_account(self, new):
        self.current_account_index = self.accounts.index(new)

    @simplegetter
    def auth_cache(self):
        return SQLiteAuthentiCache(
            directory=XDGPackageDirectory(
                "XDG_CACHE_HOME", packagename=self.name
            ).path
        )

    @simplegetter
    def request_cache(self):
        return None

    @request_cache.setter
    def request_cache(self, new):
        self._request_cache = new
        for account in self.accounts:
            account.request_cache = self.request_cache

    @property
    def config_file(self):
        with XDGPackageDirectory(
            "XDG_CONFIG_HOME", packagename=self.sanitized_name
        ) as configdir:
            return os.path.join(
                configdir, "{}.conf".format(self.sanitized_name)
            )

    def read_config(self):
        try:
            with open(self.config_file) as fh:
                logger.debug(
                    "Reading from configuration file '{}'".format(
                        self.config_file
                    )
                )
                self.config.read_file(fh)
        except (FileNotFoundError):
            logger.info(
                "Configuration file '{}' does not exist.".format(
                    self.config_file
                )
            )
            return False
        del self.accounts[:]
        for section_name in self.config:
            if "account" in section_name:
                section = self.config[section_name]
                account = SenseMapAccount(
                    email=section.get("email") or None,
                    username=section.get("username") or None,
                    api=section.get("api") or None,
                    cache_password=section.getboolean("cache_password"),
                    auth_cache=self.auth_cache
                    if section.getboolean("auth_cache")
                    else None,
                )
                self.accounts.append(account)
        self.current_account_index = self.app_config.getint(
            "current_account_index"
        )

    def write_config(self):
        # remove all account sections from the configuration
        for section_name in list(self.config):
            if "account" in section_name:
                self.config.remove_section(section_name)
        # fill the account sections
        nr = itertools.count(0)
        for account in self.accounts:
            section_name = "account:{}".format(next(nr))
            self.config.add_section(section_name)
            section = self.config[section_name]
            for option in (
                "email",
                "username",
                "api",
                "auth_cache",
                "cache_password",
            ):
                val = getattr(account, option)
                if val is None:
                    continue
                elif isinstance(val, bool) or not isinstance(val, str):
                    section[option] = "yes" if bool(val) else "no"
                elif val:
                    section[option] = str(val)
        # write the config
        with open(self.config_file, "w") as fh:
            logger.debug(
                "Writing configuration file '{}'".format(self.config_file)
            )
            self.config.write(fh)
        return True

    def __del__(self):
        self.write_config()



# system modules
import collections
import collections.abc
import datetime
import inspect
import itertools
import logging

from sensemapi.reprobject import ReprObject
from sensemapi.utils import *

# external modules


class senseBoxSensor(ReprObject):
    """
    Class representation of a senseBox sensor
    """

    def __init__(
        self,
        id=None,
        type=None,
        title=None,
        unit=None,
        icon=None,
        last_value=None,
        last_time=None,
    ):
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)

    @simplegetter
    def id(self):
        return None

    @simplesetter(id)
    def id(self, new):
        return new

    @simplegetter
    def type(self):
        return None

    @simplesetter(type)
    def type(self, new):
        return new

    @simplegetter
    def title(self):
        return None

    @simplesetter(title)
    def title(self, new):
        return new

    @simplegetter
    def unit(self):
        return None

    @simplesetter(unit)
    def unit(self, new):
        return new

    @simplegetter
    def icon(self):
        return None

    @simplesetter(icon)
    def icon(self, new):
        return new

    @simplegetter
    def last_value(self):
        return None

    @simplesetter(last_value)
    def last_value(self, new):
        return new

    @simplegetter
    def last_time(self):
        return None

    @simplesetter(last_time)
    def last_time(self, new):
        return (
            new
            if (isinstance(new, datetime.datetime) or new is None)
            else str2date(new)
        )

    @simplegetter
    def box(self):
        """
        The :class:`senseBox` this :class:`senseBoxSensor` belongs to.

        :type: :class:`senseBox`
        """
        return None

    @simplesetter(box)
    def box(self, new):
        return new

    def clear_metadata(self):
        """
        Set this sensors's metadata to :any:`None`. When uploading, this causes
        the sensor to be deleted.
        """
        for attr in (
            "title",
            "type",
            "unit",
            "icon",
            "last_time",
            "last_value",
        ):
            setattr(self, attr, None)

    def delete(self, really=False):
        """
        Delete this :class:`senseBoxSensor`.

        Args:
            really (bool, optional): Really delete this sensor?

        Returns:

        """
        assert self.id, "this sensor does not have an id"
        assert really, (
            "Refusing to delete sensor '{}' " "without really=True"
        ).format(self.id)
        assert (
            self.box
        ), "this sensor does not know what senseBox it belongs to"
        self.clear_metadata()
        return self.box.upload_metadata()

    def to_json_fallback(self, x):
        if isinstance(x, datetime.datetime):
            return date2str(x)
        return super().to_json_fallback(x)

    def update_from_api_json(self, d):
        """
        Update this :any:`senseBoxSensor` from a :any:`dict`

        Args:
            d (dict): the dictionary to update this :any:`senseBoxSensor` from
        """
        last_measurement = d.get("lastMeasurement", {}) or {}
        last_value = last_measurement.get("value")
        last_time = last_measurement.get("createdAt")
        self.id = d.get("_id")
        self.type = d.get("sensorType")
        self.unit = d.get("unit")
        self.title = d.get("title")
        self.icon = d.get("icon")
        self.last_value = float(last_value) if last_value is not None else None
        self.last_time = str2date(last_time) if last_time is not None else None

    @classmethod
    def from_api_json(cls, d):
        """
        Create a :any:`senseBoxSensor` from a :any:`dict`.

        Args:
            d (dict): the dictionary to create the :any:`senseBoxSensor` from

        Returns:
            senseBoxSensor : the senseBox sensor
        """
        sensor = cls()
        sensor.update_from_api_json(d)
        return sensor

    def to_api_json(
        self, with_id=None, edited=False, new=False, deleted=False
    ):
        """
        Serialize the metadata of this :any:`senseBoxSensor` to a JSON dict
        for the API

        Args:
            with_id (bool, optional) : serialize the ids as well? Defaults to
                including the id if it is specified.
            edited (bool, optional): add a key to mark this sensor as edited?
            new (bool, optinal): add a key to mark this sensor as new?
            deleted (bool, optional): add a key to mark this sensor as deleted?

        Returns:
            dict : the JSON dict

        Raises:
            AssertionError : if arguments conflict or data is missing
        """
        d = {}
        if new:
            edited = True
            with_id = False
        elif edited:
            with_id = True
        if deleted:
            with_id = True
        assert not (
            deleted and edited
        ), "either 'deleted' OR 'edited'/'new' can be specified"
        if with_id is None:
            if self.id is not None:
                d["_id"] = self.id
        elif with_id:
            assert self.id is not None, "no sensor id specified"
            d["_id"] = self.id
        if edited:
            assert self.title is not None, "editing, but no sensor title given"
            assert self.type is not None, "editing, but no sensor type given"
            assert self.unit is not None, "editing, but no sensor unit given"
            assert self.icon is not None, "editing, but no sensor icon given"
            d["edited"] = True
        d["title"] = self.title
        d["sensorType"] = self.type
        d["unit"] = self.unit
        d["icon"] = self.icon
        if new:
            d["new"] = True
        if deleted:
            d["deleted"] = True
        return d

    def upload_measurement(self):
        """
        Upload the current measurement. Calls
        :any:`SenseMapClient.post_measurement`.
        """
        assert self.id, "the given sensor does not have an id"
        assert self.box, "the given sensor does not know its senseBox"
        assert self.box.id, "the given sensor's senseBox does not have an id"
        assert (
            self.last_value is not None
        ), "the given sensor does not have a last_value"
        assert (
            self.box.client
        ), "This sensor's senseBox does not have an associated client"
        post_kwargs = {}
        post_kwargs.update(
            box_id=self.box.id, sensor_id=self.id, value=float(self.last_value)
        )
        if self.box.exposure == "mobile":
            post_kwargs.update(
                lat=self.box.current_lat,
                lon=self.box.current_lon,
                height=self.box.current_height,
            )
        if self.last_time:
            post_kwargs.update(time=self.last_time)
        return self.box.client.post_measurement(**post_kwargs)

    def get_measurements(self, **kwargs):
        """
        Retrieve the 10000 latest measurements of this sensor

        Args:
            kwargs: arguments passed to :meth:`SenseMapClient.get_measurements`

        Returns:
            senseBoxSensorData : the retrieved data
        """
        assert self.id, "This sensor does not have an id"
        assert self.box, "This sensor does not know its senseBox"
        assert self.box.id, "This sensor's senseBox does not have an id"
        assert (
            self.box.client
        ), "This sensor's senseBox does not have an associated client"
        get_kwargs = kwargs.copy()
        get_kwargs.update(
            box_id=self.box.id,
            sensor_id=self.id,
            format="csv",
            delimiter="comma",
        )
        data = self.box.client.get_measurements(**get_kwargs)
        data.sensor = self
        return data

    def delete_measurements(self, **kwargs):
        """
        Delete this sensor's measurements

        Args:
            kwargs: arguments passed to
                :meth:`SenseMapAccount.delete_sensor_measurements`
        """
        assert self.id, "This sensor does not have an id"
        assert self.box, "This sensor does not know its senseBox"
        assert self.box.id, "This sensor's senseBox does not have an id"
        assert (
            self.box.client
        ), "This sensor's senseBox does not have an associated client"
        assert hasattr(
            self.box.client, "delete_sensor_measurements"
        ), "This sensor's box's client cannot delete sensor measurements"
        del_kwargs = kwargs.copy()
        del_kwargs.update({"box_id": self.box.id, "sensor_id": self.id})
        return self.box.client.delete_sensor_measurements(**del_kwargs)

    def __eq__(self, other):
        """
        Compare this :class:`senseBoxSensor` to another :class:`senseBoxsensor`
        and check if all metadata is equal.
        """
        for arg in inspect.getfullargspec(self.__init__)[0][1:]:
            if not getattr(self, arg) == getattr(other, arg):
                return False
        return True


class senseBoxSensorCollection(collections.abc.MutableSequence, ReprObject):
    """
    Container for multiple :any:`senseBoxSensor` s

    Args:
        sensors (sequence, optional): the :any:`senseBoxSensor` s
    """

    def __init__(self, sensors=[]):
        self.sensors = sensors

    @property
    def sensors(self):
        """
        The underlying :any:`senseBoxSensor` s

        :type: any:`list`
        :setter: also adds the :any:`senseBoxSensorCollection.box` to the
            given :any:`senseBoxSensor` s
        """
        try:
            self._sensors
        except AttributeError:
            self._sensors = []
        return self._sensors

    @sensors.setter
    def sensors(self, new_sensors):
        self._sensors = []
        for sensor in new_sensors:
            sensor.box = self.box
            self.append(sensor)

    @simplegetter
    def box(self):
        """
        The corresponding :class:`senseBox` these :class:`senseBoxSensor` s
        belong to.

        :type: :class:`senseBox` or :any:`None`
        """
        return None

    @simplesetter(box)
    def box(self, new):
        return new

    @property
    def by_id(self):
        """
        A :any:`dict` of these :any:`senseBoxSensor` s by their
        ids. Sensors without an id get assigned to counter key
        ``new_sensor_N``.

        :type: :any:`dict`
        """
        i = itertools.count(0)
        return {
            sensor.id
            if sensor.id
            else ("new_sensor_{}".format(next(i))): sensor
            for sensor in self
        }

    @property
    def by_title(self):
        """
        A :any:`dict` of these :any:`senseBoxSensor` s by their titles. Sensors
        without a title get assigned to counter key ``unnamed sensor N``.
        Duplicate names get a ``" Nr. N"`` appended.

        :type: :any:`dict`
        """
        unnamed = itertools.count(0)
        d = {}
        duplicates = collections.Counter()
        for sensor in self:
            if sensor.title:
                if sensor.title in d:
                    duplicates[sensor.title] += 1
                    d[
                        "{} Nr. {}".format(
                            sensor.title, duplicates[sensor.title] + 1
                        )
                    ] = sensor
                else:
                    d[sensor.title] = sensor
            else:
                d["unnamed sensor {}".format(next(unnamed))] = sensor
        return d

    # boring definitions of methods
    # I tried to use a decorator for this (https://gitlab.com/snippets/1742907)
    # but unfortunately it does not work exactly as it should
    def __getitem__(self, *args, **kwargs):
        return self.sensors.__getitem__(*args, **kwargs)

    def __setitem__(self, index, value):
        """
        Also adds the :any:`senseBoxSensorCollection.box` to the new sensor
        """
        value.box = self.box
        return self.sensors.__setitem__(index, value)

    def insert(self, index, value):
        """
        Also adds the :any:`senseBoxSensorCollection.box` to the new sensor
        """
        value.box = self.box
        return self.sensors.insert(index, value)

    def append(self, value):
        """
        Also adds the :any:`senseBoxSensorCollection.box` to the new sensor
        """
        value.box = self.box
        return self.sensors.append(value)

    def __delitem__(self, *args, **kwargs):
        return self.sensors.__delitem__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        return self.sensors.__len__(*args, **kwargs)

    def __eq__(self, other):
        return self.sensors == other.sensors

    def __getattr__(self, attr):
        """
        Retreives a box by looking its title up in
        :any:`senseBoxSensorCollection.by_title`.
        """
        try:
            return self.by_title[attr]
        except KeyError:
            raise AttributeError(
                "{} object does neither have an attribute '{}' "
                "nor a box with that title".format(
                    self.__class__.__name__, attr
                )
            )



# system modules
import collections
import csv
import datetime
import functools
import inspect
import json
import logging
import os
import re

from sensemapi.errors import *

# external modules

logger = logging.getLogger(__name__)


def catch(exceptions, f, *args, **kwargs):
    """
    Execute a callable, catch specific exceptions and return value accordingly

    Args:
        exceptions (dict): mapping of exceptions to return values
        f (callable): the function to call
        args, kwargs: arguments passed to the function

    Returns:
        the return value of ``f`` or the proper value in ``exceptions`` if an
        exception was raised
    """
    try:
        return f(*args, **kwargs)
    except BaseException as e:
        for exception, retval in exceptions.items():
            if issubclass(type(e), exception):
                return retval
        raise


OPENSENSEMAP_DATETIME_FORMAT_UTC_MS = "%Y-%m-%dT%H:%M:%S.%fZ"
OPENSENSEMAP_DATETIME_FORMAT_UTC = "%Y-%m-%dT%H:%M:%SZ"


def unique(iterable):
    """
    Generator yielding only unique elements

    Args:
        iterable (iterable): iterable of :any:`hash`-able objects
    """
    seen = set()
    for x in iterable:
        if x in seen:
            continue
        else:
            seen.add(x)
            yield x


def str2date(s):
    """
    Convert a string in OpenSenseMap-RFC3339 format to a Python
    :any:`datetime.datetime` object.

    Args:
        s (str) : the string to convert. Needs to be in OpenSenseMap-RFC3339
            format.

    Returns:
        datetime.datetime : the converted date
    """
    for fmt in [
        OPENSENSEMAP_DATETIME_FORMAT_UTC_MS,
        OPENSENSEMAP_DATETIME_FORMAT_UTC,
    ]:
        try:
            d = datetime.datetime.strptime(s, fmt)
            d = d.replace(tzinfo=datetime.timezone.utc)
            return d
        except ValueError:
            pass
    raise ValueError("Could not parse date '{}' string".format(s))


def date2str(d):
    """
    Stringify a :any:`datetime.datetime` object to RFC3339 format

    Args:
        d (datetime.datetime) : the datetime to convert

    Returns:
        str : the converted string
    """
    try:
        utc = d.astimezone(datetime.timezone.utc)
    except ValueError:
        utc = d.replace(tzinfo=datetime.timezone.utc)
    s = utc.strftime(OPENSENSEMAP_DATETIME_FORMAT_UTC_MS)
    return re.sub(
        string=s,
        pattern=r"(\.\d{3})\d{3}([a-z])$",
        repl=r"\1\2",
        flags=re.IGNORECASE,
    )


def pretty_json(d):  # pragma: no cover
    """
    Pretty-format a JSON dict

    Args:
        d (dict) : the JSON dict

    Returns:
        str : the formatted string
    """
    return json.dumps(d, sort_keys=True, indent=4)


def read_csv(f, *args, **kwargs):
    """
    Read CSV from file and return a dict

    Args:
        f (filehandle): file handle
        args, kwargs: arguments passed to :any:`csv.DictReader`

    Returns:
        dict : keys are column names, values are column data
    """
    d = {}
    csvreader = csv.DictReader(f, *args, **kwargs)

    def convert(x):
        try:
            return float(x)
        except ValueError:
            try:
                return str2date(x)
            except ValueError:
                return x

    d = dictlist_to_listdict(csvreader, fun=convert)
    return d


def write_csv(f, d, headersortkey=lambda x: x, *args, **kwargs):
    """
    Write CSV dict to file

    Args:
        f (filehandle): writeable file handle
        headersortkey (callable): ``key`` argument to :any:`sorted` to sort the
            header columns
        args, kwargs: arguments passed to :any:`csv.DictWriter`
    """
    csvwriter = csv.DictWriter(
        f, fieldnames=sorted(d.keys(), key=headersortkey), *args, **kwargs
    )
    L = listdict_to_dictlist(d)
    csvwriter.writeheader()
    for row in L:
        csvwriter.writerow(row)


def listdict_to_dictlist(d):
    """
    Convert a :any:`dict` of lists to a list of dicts

    Args:
        d (dict): the dict of lists

    Returns:
        list : list of dicts
    """
    return (dict(zip(d, t)) for t in zip(*d.values()))


def dictlist_to_listdict(l, fun=lambda x: x):
    """
    Convert a list of dicts to a dict of lists

    Args:
        l (list): the list of dicts
        fun (callable): callable to manipulate the values

    Returns:
        dict : dict of lists
    """
    d = collections.defaultdict(list)
    for e in l:
        for k, v in e.items():
            d[k].append(fun(v))
    return dict(d)


def location_dict(lat=None, lon=None, height=None):
    """
    Create a location dict from the single values

    Args:
        lat, lon, height (float, optional) : the position

    Returns:
        dict : the location dict

    Raises:
        ValueError : if not enough values are specified
    """
    if lat is not None and lon is not None:
        if height is not None:
            return {"lat": lat, "lng": lon, "height": height}
        else:
            return {"lat": lat, "lng": lon}
    else:
        raise ValueError("At least 'lat' and 'lon' need to be defined")


def simplegetter(fn):  # pragma: no cover
    """
    Property getter method decorator for easy getter setup. Getter methods
    decorated with this decorator should return the default value. When the
    property is requested, it is checked whether the internal attribute
    prefixed with ``_`` exists and if so it is returned. If not, it is set
    to the return value of the decorated method and then it is returned.

    Parameters
    ----------

    Returns
    -------

    callable
        The decorated callable
    """
    propname = fn.__name__
    attrname = "_{}".format(propname)
    try:
        inspect.getfullargspec(fn)[0]
    except KeyError:  # pragma: no cover
        raise ValueError(
            "`simplegetter` decorator can only be used for methods that take "
            "the object reference as first argument"
        )

    @property
    @functools.wraps(fn)
    def getter(self):
        try:
            return getattr(self, attrname)
        except AttributeError:
            fnval = fn(self)
            setattr(self, attrname, fnval)
        return getattr(self, attrname)

    return getter


def simplesetter(prop, del_on_exceptions=()):  # pragma: no cover
    """
    Property setter method decorator for easy setter setup. Setter methods
    decorated with this decorator should take the object reference and the new
    value as argument, modify the value as desired and return it. The return
    value will be stored in the internal attribute prefixed with
    ``_``.

    Parameters
    ----------

    prop : property
        the property
    del_on_exceptions : sequence of BaseException subclasses
        delete the internal attribute if any of these in
        :any:`exceptions` occur

    Returns
    -------

    callable
        The decorated callable
    """

    def simplesetter_decorator(fn):
        propname = fn.__name__
        attrname = "_{}".format(propname)
        try:
            inspect.getfullargspec(fn)[0]
            inspect.getfullargspec(fn)[1]
        except KeyError:  # pragma: no cover
            raise ValueError(
                "`simplegetter` decorator can only be used for methods that "
                "take the object reference as first argument and the new "
                "property value as second argument"
            )

        def setter(self, newval):
            try:
                converted = fn(self, newval)
                setattr(self, attrname, converted)
            except del_on_exceptions:
                try:
                    delattr(self, attrname)
                except AttributeError:
                    pass

        setter.__doc__ = fn.__doc__
        setter = prop.setter(setter)
        return setter

    return simplesetter_decorator


def needs_pandas(f):
    """
    Decorator for methods requiring :mod:`pandas` to be installed
    """

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            import pandas
        except ImportError:  # pragma: no cover
            raise NoPandasError
        f.__globals__["pandas"] = pandas
        return f(*args, **kwargs)

    return wrapper


def needs_cachecontrol(f):
    """
    Decorator for methods requiring :mod:`cachecontrol` to be installed
    """

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            import cachecontrol
        except ImportError:  # pragma: no cover
            raise NoCacheControl
        f.__globals__["cachecontrol"] = cachecontrol
        return f(*args, **kwargs)

    return wrapper



# system modules
import collections
import inspect
import itertools
import logging

# external modules
import requests

from sensemapi import paths
from sensemapi.errors import *

# internal modules
from sensemapi.reprobject import ReprObject
from sensemapi.sensor import senseBoxSensor, senseBoxSensorCollection
from sensemapi.utils import *

logger = logging.getLogger(__name__)


class senseBox(ReprObject):
    """
    Class representation of a senseBox
    """

    def __init__(
        self,
        id=None,
        exposure=None,
        grouptag=None,
        description=None,
        name=None,
        sensors=None,
        image=None,
        current_lat=None,
        current_lon=None,
        current_height=None,
        weblink=None,
    ):
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)

    @simplegetter
    def id(self):
        return None

    @simplesetter(id)
    def id(self, new):
        return new

    @simplegetter
    def exposure(self):
        return None

    @simplesetter(exposure)
    def exposure(self, new):
        return new

    @simplegetter
    def grouptag(self):
        return None

    @simplesetter(grouptag)
    def grouptag(self, new):
        return new

    @simplegetter
    def description(self):
        return None

    @simplesetter(description)
    def description(self, new):
        return new

    @simplegetter
    def name(self):
        return None

    @simplesetter(name)
    def name(self, new):
        return new

    @property
    def sensors(self):
        try:
            self._sensors
        except AttributeError:
            self._sensors = senseBoxSensorCollection()
            self._sensors.box = self
        return self._sensors

    @sensors.setter
    def sensors(self, new):
        if hasattr(new, "sensors"):  # is already a senseBoxSensorCollection
            self._sensors = new
        else:
            self._sensors = senseBoxSensorCollection()
            self._sensors.box = self
            self._sensors.sensors = new

    @simplegetter
    def image(self):
        return None

    @simplesetter(image)
    def image(self, new):
        return new

    @simplegetter
    def weblink(self):
        return None

    @simplesetter(weblink)
    def weblink(self, new):
        return new

    @simplegetter
    def client(self):
        return None

    @simplesetter(client)
    def client(self, new):
        return new

    @simplegetter
    def current_lat(self):
        return None

    @simplesetter(current_lat)
    def current_lat(self, new):
        return new

    @simplegetter
    def current_lon(self):
        return None

    @simplesetter(current_lon)
    def current_lon(self, new):
        return new

    @simplegetter
    def current_height(self):
        return None

    @simplesetter(current_height)
    def current_height(self, new):
        return new

    @property
    def location(self):
        """
        A location object as used by the API

        :type: :any:`dict`
        """
        return location_dict(
            self.current_lat, self.current_lon, self.current_height
        )

    def new_sensor(self, *args, **kwargs):
        """
        Add a new :class:`senseBoxSensor` to this :class:`senseBox`

        Args:
            args, kwargs : arguments passed to :meth:`senseBoxSensor.__init__`

        Returns:
            senseBoxSensor : the newly created sensor
        """
        sensor = senseBoxSensor(*args, **kwargs)
        self.add_sensor(sensor)
        return sensor

    def add_sensor(self, sensor):
        """
        Add a :class:`senseBoxSensor` to this :class:`senseBox`

        Args:
            sensor (senseBoxSensor) : the sensor to add
        """
        self.sensors.append(sensor)

    def update_from_api_json(self, d):
        """
        Create a :class:`senseBox` from a :any:`dict`.

        Args:
            d (dict): the dictionary to update this :class:`senseBox` from
        """
        locfields = ["lng", "lat", "height"]
        coords = {k: None for k in locfields}
        coords.update(
            {
                k: v
                for k, v in zip(
                    locfields,
                    d.get("currentLocation", {}).get("coordinates", []),
                )
            }
            or d.get("location", {})
        )
        for sensor_json in d.get("sensors", []):
            sensor_json_copy = sensor_json.copy()
            sensor_id = sensor_json_copy.get("_id")
            own_sensor = self.sensors.by_id.get(sensor_id)
            if own_sensor:
                own_sensor.update_from_api_json(sensor_json_copy)
            else:
                sensor_json_copy.pop("_id", None)
                self.sensors.append(senseBoxSensor.from_api_json(sensor_json))
        self.id = d.get("_id")
        for arg in inspect.getfullargspec(self.__init__)[0][1:]:
            if arg in ("sensors",):
                continue
            if arg in d:
                old_val = getattr(self, arg)
                new_val = d[arg]
                if new_val != old_val:
                    logger.debug(
                        "Changing box {} from {} to {}".format(
                            arg, repr(old_val), repr(new_val)
                        )
                    )
                    setattr(self, arg, new_val)
        self.current_lat = coords.get("lat")
        self.current_lon = coords.get("lng")
        self.current_height = coords.get("height")
        self.weblink = d.get("weblink")

    @classmethod
    def from_api_json(cls, d):
        """
        Create a :class:`senseBox` from a :any:`dict`.

        Args:
            d (dict): the dictionary to create the :class:`senseBox` from

        Returns:
            senseBox : the senseBox
        """
        senseBox = cls()
        senseBox.update_from_api_json(d)
        return senseBox

    def to_api_json(self, with_id=True):
        """
        Serialize the metadata of this senseBox to a JSON dict for the API

        Args:
            with_id (bool, optional) : serialize the ids as well? Defaults to
                ``True``.

        Returns:
            dict : the JSON dict
        """
        box_json = {}
        if with_id:
            if self.id is not None:
                box_json["_id"] = self.id
        if self.name is not None:
            box_json["name"] = self.name
        if self.grouptag is not None:
            box_json["grouptag"] = self.grouptag
        box_json["location"] = {}
        if self.current_lat is not None:
            box_json["location"]["lat"] = float(self.current_lat)
        if self.current_lon is not None:
            box_json["location"]["lng"] = float(self.current_lon)
        if self.current_height is not None:
            box_json["location"]["height"] = float(self.current_height)
        if (
            "lat" not in box_json["location"]
            or "lng" not in box_json["location"]
        ):
            box_json.pop("location", None)
        if self.description is not None:
            box_json["description"] = self.description
        if self.weblink is not None:
            box_json["weblink"] = self.weblink
        if self.exposure is not None:
            box_json["exposure"] = self.exposure
        if self.sensors is not None:
            box_json["sensors"] = []
            for sensor in self.sensors:
                if sensor.id is None:
                    sensor_json = sensor.to_api_json(new=True)
                else:
                    if all(
                        [
                            v
                            for k, v in sensor.to_api_json().items()
                            if k != "_id"
                        ]
                    ):
                        sensor_json = sensor.to_api_json(edited=True)
                    else:
                        sensor_json = sensor.to_api_json(deleted=True)
                box_json["sensors"].append(sensor_json)
        return box_json

    def upload_metadata(self):
        """
        Upload the metadata of this :class:`senseBox` to the API. Tries to call
        :any:`SenseMapAccount.update_box_metadata`.

        Returns:
            True : if the updating went well

        Raises:
            NoClientError : if this senseBox does not have an associated
                :class:`senseBoxClient` in :any:`senseBox.client`
        """
        if self.client is None:
            raise NoClientError("This box does not have an associated client.")
        if not hasattr(self.client, "update_box_metadata"):
            raise NoClientError(
                "This box' client does is not able " "to update the metadata"
            )
        return self.client.update_box_metadata(self)

    def upload_as_new(self):
        """
        Upload this :class:`senseBox` as a new senseBox. Tries to call
        :any:`SenseMapAccount.new_box`.

        Returns:
            True : if the updating went well

        Raises:
            NoClientError : if this senseBox does not have an associated
                :class:`senseBoxClient` in :class:`senseBox.client`
        """
        if self.client is None:
            raise NoClientError("This box does not have an associated client.")
        if not hasattr(self.client, "new_box"):
            raise NoClientError(
                "This box' client does is not able " "to upload a new senseBox"
            )
        return self.client.new_box(self)

    def fetch_metadata(self):
        """
        Fetch the metadata

        Raises:
            NoClientError : if this senseBox does not have an associated
                :class:`senseBoxClient` in :class:`senseBox.client`
        """
        if self.client is None:
            raise NoClientError("This box does not have an associated client.")
        if not hasattr(self.client, "_get_box"):
            raise NoClientError(
                "This box' client does is not able " "to fetch boxes"
            )
        assert self.id, "this senseBox does not have an id"
        box_json = self.client._get_box(self.id)
        self.update_from_api_json(box_json)

    def delete(self, really=False):
        """
        Delete this senseBox and all of its sensors and measurements.

        Args:
            really (bool): Really delete this box? Defaults to ``False``.
        """
        if self.client is None:
            raise NoClientError("This box does not have an associated client.")
        if not hasattr(self.client, "delete_box"):
            raise NoClientError(
                "This box' client does is not able " "to delete boxes"
            )
        assert self.id, "this senseBox does not have an id"
        return self.client.delete_box(self.id, really=really)

    def __eq__(self, other):
        """
        Compare this :class:`senseBox` to another :class:`senseBox` and check
        if all metadata is equal.
        """
        for arg in inspect.getfullargspec(self.__init__)[0][1:]:
            if not getattr(self, arg) == getattr(other, arg):
                return False
        return True


class senseBoxCollection(collections.abc.MutableSequence, ReprObject):
    """
    Container for multiple :class:`senseBox` s

    Args:
        boxes (sequence, optional): the :class:`senseBox` s
    """

    def __init__(self, boxes=[]):
        self.boxes = boxes

    @property
    def boxes(self):
        """
        The underlying :class:`senseBox` es

        :type: :any:`list`
        :setter: also adds the :meth:`senseBoxCollection.client` to the
            given :class:`senseBox` s
        """
        try:
            self._boxes
        except AttributeError:
            self._boxes = []
        return self._boxes

    @boxes.setter
    def boxes(self, new_boxes):
        self._boxes = []
        for box in new_boxes:
            box.client = self.client
            self.append(box)

    @simplegetter
    def client(self):
        """
        The :class:`SenseMapClient` these :class:`senseBox` es belong to.

        :type: :class:`SenseMapClient` or :any:`None`
        """
        return None

    @simplesetter(client)
    def client(self, new):
        return new

    @property
    def by_id(self):
        """
        A :any:`dict` of these :class:`senseBox` es by their
        ids. boxes without an id get assigned to counter key
        ``new_sensor_N``.

        :type: :any:`dict`
        """
        i = itertools.count(0)
        return {
            box.id if box.id else ("new_box_{}".format(next(i))): box
            for box in self
        }

    @property
    def by_name(self):
        """
        A :any:`dict` of these :class:`senseBox` es by their names. Boxes
        without a name get assigned to counter key ``unnamed box N``.
        Duplicate names get a ``" Nr. N"`` appended.

        :type: :any:`dict`
        """
        unnamed = itertools.count(0)
        d = {}
        duplicates = collections.Counter()
        for box in self:
            if box.name:
                if box.name in d:
                    duplicates[box.name] += 1
                    d[
                        "{} Nr. {}".format(box.name, duplicates[box.name] + 1)
                    ] = box
                else:
                    d[box.name] = box
            else:
                d["Unnamed senseBox {}".format(next(unnamed))] = box
        return d

    def __getitem__(self, *args, **kwargs):
        return self.boxes.__getitem__(*args, **kwargs)

    def __setitem__(self, index, value):
        """
        Also adds the :any:`senseBoxCollection.client` to the new box
        """
        value.client = self.client
        return self.boxes.__setitem__(index, value)

    def insert(self, index, value):
        """
        Also adds the :any:`senseBoxCollection.client` to the new box
        """
        value.client = self.client
        return self.boxes.insert(index, value)

    def append(self, value):
        """
        Also adds the :any:`senseBoxCollection.client` to the new box
        """
        value.client = self.client
        return self.boxes.append(value)

    def __delitem__(self, *args, **kwargs):
        return self.boxes.__delitem__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        return self.boxes.__len__(*args, **kwargs)

    def __eq__(self, other):
        return self.boxes == other.boxes

    def __getattr__(self, attr):
        """
        Retreives a box by looking its title up in
        :any:`senseBoxCollection.by_name`.
        """
        try:
            return self.by_name[attr]
        except KeyError:
            raise AttributeError(
                "{} object does neither have an attribute '{}' "
                "nor a box with that title".format(
                    self.__class__.__name__, attr
                )
            )



# system modules
import csv
import inspect
import io
import json
import logging

from sensemapi.errors import *
from sensemapi.reprobject import ReprObject
from sensemapi.utils import *

# external modules

logger = logging.getLogger(__name__)


class senseBoxSensorData(ReprObject):
    """
    Class holding data of a :class:`senseBoxSensor`

    Args:
        data (dict or str, optional): the data data. Either a CSV string or a
            :any:`dict` as returned by the API when downloading timeseries.
    """

    def __init__(self, data=None, sensor=None):
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)

    @simplegetter
    def data(self):
        """
        The data data

        :type: :any:`str` or :any:`dict`
        """

    @simplesetter(data)
    def data(self, new):
        return new

    @simplegetter
    def sensor(self):
        """
        The this data belongs to sensor

        :type: :any:`senseBoxSensor`
        """

    @simplesetter(sensor)
    def sensor(self, new):
        return new

    @classmethod
    def from_csv(cls, fh, **kwargs):
        """
        Create a :class:`senseBoxSensorData` object from a CSV string

        Args:
            fh (file-like object): file-like object (e.g. :class:`io.StringIO`)
            kwargs : further keyword arguments passed to
                :meth:`senseBoxSensorData.__init__`.

        Returns:
            senseBoxSensorData
        """
        data = read_csv(fh)
        if not data:
            data = {"createdAt": [], "value": []}
        assert len(data) == 2, "data does not have exactly two columns"
        for col in ("createdAt", "value"):
            assert col in data, "data does not have '{}' column".format(col)
        return cls(data=data)

    @property
    def csv(self):  # pragma: no cover
        """
        Serializes this data to CSV. **Warning**: This is **not** consistent
        with :meth:`senseBoxSensorData.from_csv`.

        Returns:
            str : the CSV data
        """
        csvfh = io.StringIO()
        write_csv(csvfh, self.data)
        return csvfh.getvalue()

    @property
    @needs_pandas
    def series(self):
        """
        Convert the data to a :class:`pandas.Series`

        Returns:
            pandas.Series : the data as series
        """
        series = pandas.Series(
            data=self.data["value"], index=self.data["createdAt"]
        )
        if self.sensor:
            series.name = "{} ({}) [{}]".format(
                self.sensor.title, self.sensor.type, self.sensor.unit
            )
        return series

    def __len__(self):
        """
        Return the number of data rows

        Raises:
            ValueError : If the lengths are strange
        """
        try:
            l1 = len(self.data["createdAt"])
            l2 = len(self.data["value"])
        except (KeyError, TypeError) as e:  # pragma: no cover
            raise ValueError("data is ill-formed: {}".format(e))
        if l1 == l2:
            return l1
        else:  # pragma: no cover
            raise ValueError(
                "'createdAt' has {} elements "
                "and 'value' has {} elements".format(l1, l2)
            )



# system modules
import collections
import datetime
import functools
import glob
import hashlib
import inspect
import itertools
import json
import logging
import os
import re
import time
from contextlib import contextmanager
from posixpath import join as urljoin

# external modules
import requests

# internal modules
from sensemapi import compat, paths
from sensemapi.client import SenseMapClient
from sensemapi.errors import *
from sensemapi.senseBox import senseBox, senseBoxCollection
from sensemapi.utils import *
from sensemapi.xdg import *

logger = logging.getLogger(__name__)


class SenseMapAccount(SenseMapClient):
    """
    Client to interface an `OpenSenseMap <https://opensensemap.org>`_ account
    via the `OpenSenseMap API <https://api.opensensemap.org>`_

    Args:
        email (str, optional): the user's login email address
        username (str, optional): the user's login name
        password (str, optional): the user's login password
        api (str, optional): the api server to use. You may use
            :any:`OPENSENSEMAP_API_LIVE` (default) or
            :any:`OPENSENSEMAP_API_TEST` for testing purposes.
        token (str, optional): the JSON Web Token
        token_time (str, optional): retrieval time of the JSON Web Token
        refresh_token (str, optional): the refresh JSON Web Token
        refresh_token_time (str, optional): retrieval time of the refresh JSON
            Web Token
        auth_cache (AuthentiCache, optional): The authentication cache to use.
            Defaults to no cache.
        request_cache (cachecontrol.adapter.CacheControlAdapter, optional):
            request cache. Defaults to no cache.
        cache_password (bool, optional): cache the password (in plain text)?
            Defaults to ``False``. Only used if ``auth_cache`` is set.
        connection_timeout (float, optional): default connection timeout
            for requests in seconds. Default is no timeout.
        response_timeout (float, optional): default response timeout
            for requests in seconds. Default is no timeout.
    """

    def __init__(
        self,
        email=None,
        username=None,
        password=None,
        token=None,
        token_time=None,
        refresh_token=None,
        refresh_token_time=None,
        api=None,
        cache_password=None,
        auth_cache=None,
        request_cache=None,
        connection_timeout=None,
        response_timeout=None,
    ):
        frame = inspect.currentframe()
        args = inspect.getargvalues(frame)[0]
        for arg in args[1:]:
            val = locals().get(arg)
            if val is not None:
                setattr(self, arg, val)

    def sync_auth_cache_afterwards(decorated_function):
        """
        Decorator for methods to call :meth:`sync_auth_cache` after
        completition
        """

        @functools.wraps(decorated_function)
        def wrapper(self, *args, **kwargs):
            ret = decorated_function(self, *args, **kwargs)
            if self.auth_cache:
                logger.debug(
                    "calling sync_auth_cache after {}".format(
                        decorated_function
                    )
                )
                self.sync_auth_cache()
            return ret

        return wrapper

    @simplegetter
    def api(self):
        return paths.OPENSENSEMAP_API_LIVE

    @api.setter
    @sync_auth_cache_afterwards
    @SenseMapClient.modifies_session
    def api(self, new):
        self._api = new

    @simplegetter
    def email(self):
        return None

    @email.setter
    @sync_auth_cache_afterwards
    def email(self, new):
        self._email = new

    @simplegetter
    def username(self):
        return None

    @username.setter
    @sync_auth_cache_afterwards
    def username(self, new):
        self._username = new

    @simplegetter
    def password(self):
        return None

    @password.setter
    @sync_auth_cache_afterwards
    def password(self, new):
        self._password = new

    @simplegetter
    def role(self):
        return None

    @simplesetter(role)
    def role(self, new):
        return new

    @simplegetter
    def language(self):
        return None

    @simplesetter(language)
    def language(self, new):
        return new

    @simplegetter
    def application(self):
        return None

    @application.setter
    def application(self, app):
        self._application = app

    @simplegetter
    def auth_cache(self):
        return None

    @auth_cache.setter
    @sync_auth_cache_afterwards
    def auth_cache(self, new):
        if new:
            if all(map(lambda x: hasattr(new, x), ("get", "sync"))):
                self._cache = new
            else:
                if self.application:
                    if self.application.auth_cache:
                        self._cache = self.application.auth_cache
                    else:
                        raise ValueError(
                            "{} object is not suitable as cache and "
                            "this account's associated application does not "
                            "have a cache".format(type(new).__name__).format(
                                new
                            )
                        )
                else:
                    raise ValueError(
                        "{} object is not suitable as cache and "
                        "this account does not have an associated "
                        "application".format(type(new).__name__).format(new)
                    )
        self._auth_cache = new

    @property
    def request_cache(self):
        """
        The request cache for this account.

        :getter: if a cache has already been set on this account, return it.
            Otherwise, return this account's application's
            :meth:`SenseMapApplication.request_cache` if available.
        """
        try:
            return self._request_cache
        except AttributeError:
            if self.application:
                return self.application.request_cache
            else:
                return None

    @request_cache.setter
    def request_cache(self, new):
        self._request_cache = new

    @simplegetter
    def cache_password(self):
        return False

    @cache_password.setter
    @sync_auth_cache_afterwards
    def cache_password(self, new):
        self._cache_password = bool(new)

    @property
    def boxes(self):
        try:
            self._boxes
        except AttributeError:
            self._boxes = senseBoxCollection()
            self._boxes.client = self
        return self._boxes

    @boxes.setter
    def boxes(self, new):  # pragma: no cover
        if hasattr(new, "boxes"):  # is already a senseBoxCollection
            self._boxes = new
        else:
            self._boxes = senseBoxCollection()
            self._boxes.client = self
            self._boxes.boxes = new

    @simplegetter
    def token(self):
        return None

    @simplesetter(token)
    def token(self, new):
        if not new:
            self.token_time = None
        return new

    @simplegetter
    def refresh_token(self):
        return None

    @simplesetter(refresh_token)
    def refresh_token(self, new):
        if not new:
            self.refresh_token_time = None
        return new

    @simplegetter
    def token_time(self):
        return None

    @simplesetter(token_time)
    def token_time(self, new):
        return new

    @simplegetter
    def refresh_token_time(self):
        return None

    @simplesetter(refresh_token_time)
    def refresh_token_time(self, new):
        return new

    @property
    def authorization_header(self):
        """
        Authorization header

        :type: :any:`dict`
        :getter: Return a :mod:`requests`-compatible header-dict containing the
            authorization token. :any:`sign_in` if necessary.
        """
        if not self.token:
            self.sign_in()
        return {"Authorization": "Bearer {}".format(self.token)}

    @property
    def authenticache(self):
        """
        The authentication :any:`dict` to cache

        :type: :any:`dict`
        """
        authenticache = {}
        for attr in (
            "token",
            "refresh_token",
            "token_time",
            "refresh_token_time",
            "password",
            "username",
            "email",
            "api",
        ):
            value = getattr(self, attr)
            authenticache.update({attr: value})
        if not self.cache_password:
            authenticache["password"] = None
        elif not self.password:
            authenticache.pop("password", None)
        return authenticache

    @contextmanager
    def pause_caching(self):
        """
        Context manager for pausing the caching temporarily
        """
        auth_cache = self.auth_cache
        if self.auth_cache:
            logger.debug("Pausing the cache temporarily")
            self.auth_cache = None
        yield
        if self.auth_cache is not auth_cache:
            self.auth_cache = auth_cache

    def delete_cache(self):
        """
        Delete the cache for this account
        """
        login = {
            attr: getattr(self, attr)
            for attr in ("email", "username", "api")
            if getattr(self, attr)
        }
        full_cache = self.authenticache
        full_cache.update(login)
        if self.auth_cache:
            self.auth_cache.delete(full_cache)

    def sync_auth_cache(self):
        """
        Synchronize with the :any:`auth_cache`
        """
        if not self.auth_cache:
            return False
        login = {
            attr: getattr(self, attr)
            for attr in ("email", "username", "api")
            if getattr(self, attr)
        }
        matching = self.auth_cache.get(conditions=login)
        our_cache = self.authenticache
        our_cache.update(login)
        updated = self.auth_cache.sync(our_cache)
        for k, v in updated.items():
            val = getattr(self, k)
            if val != v:
                if k == "password" and not self.cache_password:
                    continue
                if not v and val:
                    logger.debug("Emptying the {} due to cache".format(k))
                setattr(self, k, v)

    def request(self, *args, **kwargs):
        """
        Wrapper around :any:`SenseMapClient.request` that handles some more
        corner cases

        Raises:
            OpenSenseMapAPIOutdatedTokenError : if the tokens are outdated
            OpenSenseMapAPIAuthenticationError : if access is prohibited
        """
        response = super().request(*args, **kwargs)
        # forbidden
        if response.status_code == 403:
            try:
                response_json = response.json()
            except compat.JSONDecodeError:  # pragma: no cover
                response_json = {}
            code = response_json.get("code", "")
            message = response_json.get("message", "")
            if re.search("invalid.+jwt", message.lower()) or re.search(
                r"refresh\s+token.+invalid.+too.+old", message.lower()
            ):
                raise OpenSenseMapAPIOutdatedTokenError(
                    (code + ": " if code else "" + message)
                )
            if re.search(r"user.+password.+not\s+valid", message.lower()):
                raise OpenSenseMapAPIInvalidCredentialsError(message)
            # something else went wrong
            raise OpenSenseMapAPIAuthenticationError(
                (code + ": " if code else "" + message)
            )  # pragma: no cover
        return response

    @sync_auth_cache_afterwards
    def sign_in(self):
        """
        Sign in with the user's credentials.

        Returns:
            True : if the login process was successful

        Raises:
            NoEmailError: if no email is set
            NoPasswordError: if no password is set
            OpenSenseMapAPIInvalidCredentialsError : wrong credentials
            OpenSenseMapAPIAuthenticationError : if the login did not work
        """
        if self.email is None and self.username is None:
            raise NoUserError("Cannot sign_in: no username or email specified")
        if self.password is None:
            raise NoPasswordError("Cannot sign in: no user password specified")
        response = self.request(
            "post",
            urljoin(self.api, paths.SIGN_IN),
            json={
                "email": self.email or self.username,
                "password": self.password,
            },
        )
        response_json = response.json()
        try:
            message = response_json["message"]
            self.token = response_json["token"]
            self.token_time = datetime.datetime.utcnow()
            self.refresh_token = response_json["refreshToken"]
            self.refresh_token_time = datetime.datetime.utcnow()
            data = response_json["data"]
            user = data["user"]
        except KeyError as e:  # pragma: no cover
            raise OpenSenseMapAPIResponseError(
                "API response did not contain {}".format(e)
            )
        self.read_user_details(user)
        return True

    @sync_auth_cache_afterwards
    def sign_out(self):
        """
        Sign out the user

        Returns:
            True : if the logout process was successful

        Raises:
            OpenSenseMapAPIOutdatedTokenError : if the logout did not work
        """
        if self.token:
            response = self.request(
                "post",
                urljoin(self.api, paths.SIGN_OUT),
                headers=self.authorization_header,
            )
            response_json = response.json()
            try:
                code = response_json["code"]
                message = response_json["message"]
                if code != "Ok":  # pragma: no cover
                    raise OpenSenseMapAPIError(
                        "Logout did not work" + (": " + message or "")
                    )
            except KeyError as e:  # pragma: no cover
                raise OpenSenseMapAPIResponseError(
                    "API response did not contain {}".format(e)
                )
        self.token = None
        self.refresh_token = None
        self.token_time = datetime.datetime.now()
        self.refresh_token_time = datetime.datetime.utcnow()
        return True

    def refresh_tokens(self):
        """
        :any:`_refresh_tokens` and :any:`sign_in` again if necessary.

        Returns:
            True : if the refreshing was successful

        Raises:
            OpenSenseMapAPIAuthenticationError : if refreshing didn't work
        """
        try:
            self._refresh_tokens()
        except OpenSenseMapAPIOutdatedTokenError:
            logger.debug("Tokens are outdated. Signing in...")
            self.sign_in()
        return True

    @sync_auth_cache_afterwards
    def _refresh_tokens(self):
        """
        Attempt to refresh the tokens

        Returns:
            True : if the refreshing was successful

        Raises:
            OpenSenseMapAPIOutdatedTokenError : the refresh token is outdated
        """
        assert (
            self.refresh_token
        ), "No refresh token available. Please sign_in()"
        response = self.request(
            "post",
            urljoin(self.api, paths.REFRESH),
            headers=self.authorization_header,
            json={"token": self.refresh_token},
        )
        response_json = response.json()
        try:
            code = response_json["code"]
            message = response_json["message"]
            if code == "Forbidden":  # pragma: no cover
                if re.search("token.+invalid.+too.+old", message.lower()):
                    raise OpenSenseMapAPIOutdatedTokenError(
                        message + ". You should actually never see this."
                    )
            if code != "Authorized":  # pragma: no cover
                raise OpenSenseMapAPIError(
                    "Refreshing tokens did not work" + (": " + message or "")
                )
            self.token = response_json["token"]
            self.token_time = datetime.datetime.utcnow()
            self.refresh_token = response_json["refreshToken"]
            self.refresh_token_time = datetime.datetime.utcnow()
        except KeyError as e:  # pragma: no cover
            raise OpenSenseMapAPIResponseError(
                "API response did not contain {}".format(e)
            )
        return True

    def refresh_and_retry_on_outdated_tokens(decorated_function):
        """
        Decorator to call a decorated method, catching an
        :any:`OpenSenseMapAPIOutdatedTokenError`, then doing
        :any:`refresh_tokens` and then retries.
        """

        @functools.wraps(decorated_function)
        def wrapper(self, *args, **kwargs):
            try:
                return decorated_function(self, *args, **kwargs)
            except OpenSenseMapAPIOutdatedTokenError:
                logger.debug("Tokens are outdated. Refreshing tokens...")
                self.refresh_tokens()
            logger.debug(
                "Now trying again to call {}".format(decorated_function)
            )
            return decorated_function(self, *args, **kwargs)

        return wrapper

    def get_box(self, id):
        """
        Call :any:`SenseMapClient.get_box` and update own boxes if present

        Args:
            id (str) : the senseBox id to retreive

        Returns:
            senseBox : the retreived senseBox
        """
        box = super().get_box(id)
        logger.debug("Downloaded box '{}':\n{}".format(box.id, box))
        try:
            own_box = self.boxes.by_id[box.id]
            logger.debug(
                "We already have box '{}', updating it!".format(box.id)
            )
            own_box.update_from_api_json(box.to_api_json())
            for sensor, own_sensor in zip(box.sensors, own_box.sensors):
                own_sensor.update_from_json(sensor.to_json())
        except KeyError:  # pragma: no cover
            pass
        return box

    def get_own_boxes(self):
        """
        Run :any:`get_details` to fetch the current box ids and then
        call :any:`get_box` on all of them.

        Returns:
            senseBoxCollection : all own boxes
        """
        self.get_details()
        for box in self.boxes:
            self.get_box(box.id)
        return self.boxes

    @refresh_and_retry_on_outdated_tokens
    def get_details(self):
        """
        Get the user details

        Returns:
            True : if the retrieval was successful

        Raises:
            OpenSenseMapAPIError : if the retreival did not work
        """
        response = self.request(
            "get",
            urljoin(self.api, paths.USER_DETAILS),
            headers=self.authorization_header,
        )
        response_json = response.json()
        try:
            code = response_json["code"]
            message = response_json.get("message")
            if code != "Ok":  # pragma: no cover
                raise OpenSenseMapAPIError(
                    "Retrieving user details did not work"
                    + (": " + message or "")
                )
            data = response_json["data"]
        except KeyError as e:  # pragma: no cover
            raise OpenSenseMapAPIResponseError(
                "API response did not contain {}".format(e)
            )
        try:
            user = data["me"]
        except KeyError as e:  # pragma: no cover
            raise OpenSenseMapAPIResponseError(
                "API response data did not contain {}".format(e)
            )
        self.read_user_details(user)
        return True

    @refresh_and_retry_on_outdated_tokens
    def update_box_metadata(self, box):
        """
        Update metadata of a :class:`senseBox`

        Args:
            box (senseBox) : the box to update

        Returns:
            True : if the updating went well
        """
        box_json = box.to_api_json(with_id=True)
        logger.debug(
            "Uploading senseBox metadata:\n{}".format(pretty_json(box_json))
        )
        response = self.request(
            "put",
            urljoin(self.api, paths.BOXES, box.id),
            headers=self.authorization_header,
            json=box_json,
        )
        response_json = response.json()
        try:
            code = response_json["code"]
            message = response_json.get("message", "")
            if code != "Ok":
                raise OpenSenseMapAPIError(
                    "Could not update senseBox with id '{}'{}".format(
                        box.id, ": {}".format(message) if message else ""
                    )
                )
            data = response_json["data"]
        except KeyError as e:  # pragma: no cover
            raise OpenSenseMapAPIResponseError(
                "Response did not contain {}".format(e)
            )
        # only keep edited boxes, not deleted and new ones
        edited_boxes_ids = set(
            sensor_json.get("_id")
            for sensor_json in box_json.get("sensors", [])
            if (sensor_json.get("edited") and not sensor_json.get("new"))
        )
        box.sensors = [
            sensor for sensor in box.sensors if sensor.id in edited_boxes_ids
        ]
        # update the box data
        box.update_from_api_json(data)

    @refresh_and_retry_on_outdated_tokens
    def new_box(self, box):
        """
        Upload a new :class:`senseBox`

        Args:
            box (senseBox) : the box to upload

        Returns:
            senseBox : the responded new senseBox
        """
        assert box.exposure, "new box needs 'exposure'"
        assert box.current_lat, "new box needs 'current_lat'"
        assert box.current_lon, "new box needs 'current_lon'"
        assert box.name, "new box needs 'name'"
        assert box.sensors, "new box needs at least one sensor"
        box_json = box.to_api_json(with_id=False)
        logger.debug(
            "Uploading new senseBox:\n{}".format(pretty_json(box_json))
        )
        response = self.request(
            "post",
            urljoin(self.api, paths.BOXES),
            headers=self.authorization_header,
            json=box_json,
        )
        response_json = response.json()
        try:
            message = response_json.get("message", "")
            if "successfully" not in message:  # pragma: no cover
                raise OpenSenseMapAPIError(
                    "Could not post new senseBox{}".format(
                        ": {}".format(message) if message else ""
                    )
                )
            data = response_json["data"]
        except KeyError as e:  # pragma: no cover
            raise OpenSenseMapAPIResponseError(
                "Response did not contain {}".format(e)
            )
        data.update({"description": box.description, "image": box.image})
        box.update_from_api_json(data)
        box.client = self
        new_box = senseBox.from_api_json(data)
        logger.debug("new box:\n{}".format(new_box))
        self.get_details()
        return self.get_box(id=new_box.id)

    @refresh_and_retry_on_outdated_tokens
    def delete_box(self, box_id, really=False):
        """
        Mark a senseBox for deletion. This will also delete all the
        measurements. Deletion does not happen immediately. Also calls
        :any:`get_details`.

        Args:
            box_id (str) : the :any:`senseBox.id`
            really (bool, optional): really delete this box? Defaults to
                ``False``.

        Returns:
            True : if the marking for deletion went well
        """
        assert (
            really
        ), "Refusing to delete box '{}' without really=True".format(box_id)
        response = self.request(
            "delete",
            urljoin(self.api, paths.BOXES, box_id),
            headers=self.authorization_header,
            json={"password": self.password},
        )
        try:
            response_json = response.json()
            code = response_json.get("code", "")
            message = response_json.get("message", "")
            if re.search("user.+not.+own.+box", message.lower()):
                raise OpenSenseMapAPIPermissionError(
                    message or "User does not own box '{}'".format(box_id)
                )
            if (
                not response.status_code == requests.codes.OK
            ):  # pragma: no cover
                raise OpenSenseMapAPIError(
                    "Could not mark senseBox '{}' for deletion{}".format(
                        box_id, ": {}".format(message) if message else ""
                    )
                )
        except compat.JSONDecodeError:  # pragma: no cover
            raise OpenSenseMapAPIResponseError("Response is not JSON")
        except KeyError as e:  # pragma: no cover
            raise OpenSenseMapAPIResponseError(
                "Response did not contain {}".format(e)
            )
        self.get_details()
        return True

    @refresh_and_retry_on_outdated_tokens
    def upload_current_measurement(self, box, sensor):
        """
        Upload the current measurements of a given sensor of a given senseBox

        Args:
            box_id (str) : the :any:`senseBox.id`
            really (bool, optional): really delete this box? Defaults to
                ``False``.

        Returns:
            True : if the upload went well
        """
        assert box.id, "box has no 'id'"
        assert sensor.id, "sensor has no 'id'"
        response = self.request(
            "post",
            urljoin(self.api, paths.BOXES, box.id, sensor.id),
            headers=self.authorization_header,
            json={},
        )
        response_json = response.json()
        message = response_json.get("message", "")
        if re.search("user.+not.+own.+box", message.lower()):
            raise OpenSenseMapAPIPermissionError(
                message or "User does not own box '{}'".format(box_id)
            )
        if not response.status_code == requests.codes.OK:
            raise OpenSenseMapAPIError(
                "Could not mark senseBox '{}' for deletion{}".format(
                    box_id, ": {}".format(message) if message else ""
                )
            )
        self.get_details()
        return True

    def read_user_details(self, user_details, exception=OpenSenseMapAPIError):
        """
        Read user details into object properties

        Args:
            user_details (dict): the user details as returned by the API
            exception (Exception): the exception to raise if the data is
                ill-formed

        Returns:
            True : if everything is alright

        Raises:
            ``exception`` : if the data is ill-formed
        """
        try:
            username = user_details["name"]
            email = user_details["email"]
            if self.username != username:
                if self.username is not None:
                    logger.warning(
                        "The real username '{}' is different from "
                        "the configured username '{}'".format(
                            username, self.username
                        )
                    )
                self.username = username
            if self.email != email:
                if self.email is not None:
                    logger.warning(
                        "The real email '{}' is different from "
                        "the configured email '{}'".format(email, self.email)
                    )
                self.email = email
            if not user_details["emailIsConfirmed"]:
                logger.warning("The email '{}' is unconfirmed!".format(email))
            self.role = user_details["role"]
            self.language = user_details["language"]
            for own_box in self.boxes:
                if own_box.id not in user_details["boxes"]:
                    self.boxes.remove(own_box)
            for box_id in user_details["boxes"]:
                if box_id not in self.boxes.by_id:
                    box = senseBox(id=box_id)
                    self.boxes.append(box)
        except KeyError as e:  # pragma: no cover
            raise exception("user information did not contain {}".format(e))
        return True

    @refresh_and_retry_on_outdated_tokens
    def delete_sensor_measurements(
        self,
        box_id,
        sensor_id,
        timestamps=None,
        from_date=None,
        to_date=None,
        all=None,
        really=False,
    ):
        """
        Issue a request to delete measurements from a sensor

        Args:
            box_id (str): the senseBox id
            sensor_id (str): the sensor id
            to_date, from_date (datetime.datetime, optional): start and end
                dates to delete the data
            timestamps (list of datetime.datetime, optional): timestamps to
                delete
            all (bool, optional): delete all measurements?
            really (bool, optional): really delete the measurements?

        Returns:
            bool : whether the deletion worked
        """
        assert really, (
            "Refusing to delete measurements from sensor "
            "'{}' without really=True"
        ).format(sensor_id)
        d = {}
        if from_date is not None:
            d.update({"from-date": date2str(from_date)})
        if to_date is not None:
            d.update({"to-date": date2str(to_date)})
        if all is not None:
            d.update({"deleteAllMeasurements": bool(all)})
        if timestamps is not None:
            d.update({"timestamps": [date2str(x) for x in timestamps]})
        logger.debug("Request payload:\n{}".format(pretty_json(d)))
        response = self.request(
            "delete",
            urljoin(self.api, paths.BOXES, box_id, sensor_id, "measurements"),
            headers=self.authorization_header,
            json=d,
        )
        response_json = response.json()
        message = response_json.get("message", "")
        if not response.status_code == requests.codes.OK:  # pragma: no cover
            raise OpenSenseMapAPIError(
                "Could not delete measurements from sensor '{}'".format(box_id)
                + ": {}".format(message)
                if message
                else ""
            )
        return True

    def __del__(self):
        """
        Class deconstructor. Calls :any:`sign_out`.
        """
        if not self.auth_cache:
            try:
                self.sign_out()
            except (
                SenseMapiError,
                requests.exceptions.ConnectionError,
            ):  # pragma: no cover
                pass
        super().__del__()


class SenseMapAccountCollection(collections.abc.MutableSequence, ReprObject):
    """
    Container for multiple :any:`SenseMapAccount` s

    Args:
        accounts (sequence, optional): the :any:`SenseMapAccount` s
    """

    def __init__(self, accounts=[]):
        self.accounts = accounts

    @property
    def accounts(self):
        """
        The underlying :class:`SenseMapAccount` s

        :type: any:`list`
        :setter: also adds the :meth:`SenseMapCollection.client` to the
            given :class:`SenseMapAccount` s
        """
        try:
            self._accounts
        except AttributeError:
            self._accounts = []
        return self._accounts

    @accounts.setter
    def accounts(self, new_accounts):
        self._accounts = []
        for account in new_accounts:
            account.application = self.application
            self.append(account)

    @simplegetter
    def application(self):
        """
        The :class:`SenseMapApplication` these :class:`SenseMapAccount` s
        belong to.

        :type: :class:`SenseMapApplication` or :any:`None`
        :setter: sets the
        """
        return None

    @application.setter
    def application(self, new):
        for account in self:
            account.application = new
        self._application = new
        for account in self:
            account.application = self.application

    @property
    def by_username(self):
        """
        A :any:`dict` of these :any:`SenseMapAccount` s by their usernames.
        Accounts without a username get assigned to counter key ``account
        without username N``.
        Duplicate names get a ``" Nr. N"`` appended.

        :type: :any:`dict`
        """
        unnamed = itertools.count(0)
        d = {}
        duplicates = collections.Counter()
        for account in self:
            if account.username:
                if account.username in d:
                    duplicates[account.name] += 1
                    d[
                        "{} Nr. {}".format(
                            account.username, duplicates[account.username] + 1
                        )
                    ] = account
                else:
                    d[account.username] = account
            else:
                d[
                    "account without username {}".format(next(unnamed))
                ] = account
        return d

    def __getitem__(self, *args, **kwargs):
        return self.accounts.__getitem__(*args, **kwargs)

    def __setitem__(self, index, value):
        """
        Also adds the :any:`SenseMapAccountCollection.application` to the new
        account
        """
        value.application = self.application
        return self.accounts.__setitem__(index, value)

    def insert(self, index, value):
        """
        Also adds the :any:`SenseMapAccountCollection.application` to the new
        account
        """
        value.application = self.application
        return self.accounts.insert(index, value)

    def append(self, value):
        """
        Also adds the :any:`SenseMapAccountCollection.application` to the new
        account
        """
        value.application = self.application
        return self.accounts.append(value)

    def __delitem__(self, *args, **kwargs):
        return self.accounts.__delitem__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        return self.accounts.__len__(*args, **kwargs)

    def __eq__(self, other):
        return self.accounts == other.accounts

    def __getattr__(self, attr):
        """
        Retreives an account by looking its title up in
        :any:`SenseMapAccountCollection.by_username`.
        """
        try:
            return self.by_username[attr]
        except KeyError:
            raise AttributeError(
                "{} object does neither have an attribute '{}' "
                "nor an account with that title".format(
                    self.__class__.__name__, attr
                )
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# System modules
import os
import sys
import runpy
from setuptools import setup, find_packages


def read_file(filename):
    with open(filename, errors="ignore") as f:
        return f.read()


package = find_packages(exclude=["tests"])[0]

rpm_package = "python3-" + package
try:
    rpm = "rpm" in sys.argv[1]
except IndexError:
    rpm = False

# run setup
setup(
    name=rpm_package if rpm else package,
    description="Pythonic access to the OpenSenseMap API",
    author="Yann Büchau",
    author_email="nobodyinperson@gmx.de",
    keywords="opensensemap,sensemap,api",
    license="GPLv3",
    version=runpy.run_path(os.path.join(package, "version.py")).get(
        "__version__", "0.0.0"
    ),
    url="https://gitlab.com/tue-umphy/co2mofetten/python3-sensemapi",
    long_description=read_file("README.md"),
    long_description_content_type="text/markdown",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
    install_requires=["requests"],
    tests_require=["numpy", "pandas"],
    extras_require={
        "pandas": ["pandas"],
        "cache": ["CacheControl[filecache]>=0.12.5"],
    },
    test_suite="tests",
    packages=[package],
)



# system modules
import logging
import calendar
import datetime
from email.utils import parsedate, formatdate

# internal modules
from sensemapi.utils import *

# external modules
from cachecontrol.heuristics import BaseHeuristic
from cachecontrol.cache import BaseCache

logger = logging.getLogger(__name__)


class CacheHeuristic(BaseHeuristic):
    """
    Heuristic to mock cache-control and expires header for API responses
    """

    def update_headers(self, response):
        if "date" in response.headers:
            date = parsedate(response.headers["date"])
        else:
            logger.info("No 'date' in headers. Using current time.")
            date = datetime.datetime.utcnow().timetuple()
        expires = datetime.datetime(*date[:6]) + datetime.timedelta(days=7)
        extra_headers = {
            "expires": formatdate(calendar.timegm(expires.timetuple())),
            "cache-control": "public",
        }
        return extra_headers

    def warning(self, response):
        msg = "Automatically cached! Response might be stale."
        return "110 - '{}'".format(msg)



# system modules

# internal modules
from sensemapi.version import __version__
import sensemapi.client
import sensemapi.account
import sensemapi.sensor
import sensemapi.senseBox
import sensemapi.senseBoxSensorData
import sensemapi.utils
import sensemapi.paths
import sensemapi.xdg
import sensemapi.cache
import sensemapi.application

# external modules



# system modules
import inspect
import json
import re
import warnings

# internal modules

# external modules


class ReprObject(object):  # pragma: no cover
    """
    Simple base class that defines a :any:`__repr__` method based on an
    object's ``__init__`` arguments and properties that are named equally.
    Subclasses of :any:`ReprObject` should thus make sure to have properties
    that are named equally as their ``__init__`` arguments.
    """

    @classmethod
    def _full_variable_path(cls, var):
        """ Get the full string of a variable

        Args:
            var (any): The variable to get the full string from

        Returns:
            str : The full usable variable string including the module
        """
        if inspect.ismethod(var):  # is a method
            string = "{module}.{cls}.{name}".format(
                name=var.__name__,
                cls=var.__self__.__class__.__name__,
                module=var.__module__,
            )
        else:
            name = var.__name__
            module = var.__module__
            if module == "builtins":
                string = name
            else:
                string = "{module}.{name}".format(name=name, module=module)
        return string

    @property
    def init_arguments(self):
        init_args = {}
        for arg in inspect.getfullargspec(self.__init__).args[1:]:
            try:
                init_args[arg] = getattr(self, arg)
            except AttributeError:
                warnstr = (
                    "class {cls} has no property or attribute "
                    "'{arg}' like the argument in its __init__. "
                ).format(cls=classname, arg=arg)
                warnings.warn(warnstr)
        return init_args

    def to_json_fallback(self, x):
        try:
            return x.to_json()
        except (AttributeError):
            return str(x)

    def to_json(self):
        return json.loads(
            json.dumps(self.init_arguments, default=self.to_json_fallback)
        )

    @classmethod
    def from_json(cls, d):
        init_argspec = inspect.getfullargspec(self.__init__)
        init_args = {
            arg: default
            for arg, default in zip(
                init_argspec.args[1:], init_argspec.defaults[1:]
            )
        }
        for k, v in d.items():
            if k in init_args:
                init_args[k] = v
        obj = cls(**init_args)
        return obj

    def update_from_json(self, d):
        for k, v in d.items():
            try:
                setattr(self, k, v)
            except AttributeError:
                pass

    def __repr__(self):
        """
        Python representation of this object

        Returns:
            str : a Python representation of this object based on its
            ``__init__`` arguments and corresponding properties.
        """
        indent = "    "
        # the current "full" classname
        classname = self._full_variable_path(self.__class__)

        # get a dict of {'argname':'property value'} from init arguments
        init_args = {}  # start with empty dict
        for arg in inspect.getfullargspec(self.__init__).args[1:]:
            try:
                attr = getattr(self, arg)  # get the attribute
                try:
                    string = self._full_variable_path(attr)
                except BaseException:
                    string = repr(attr)

                # indent the arguments
                init_args[arg] = re.sub(
                    string=string, pattern="\n", repl="\n" + indent
                )
            except AttributeError:  # no such attribute
                warnstr = (
                    "class {cls} has no property or attribute "
                    "'{arg}' like the argument in its __init__. "
                    "Cannot include argument '{arg}' into __repr__."
                ).format(cls=classname, arg=arg)
                warnings.warn(warnstr)

        # create "arg = {arg}" string list for reprformat
        args_kv = []
        for arg in init_args.keys():
            args_kv.append(indent + "{arg} = {{{arg}}}".format(arg=arg))

        # create the format string
        if args_kv:  # if there are arguments
            reprformatstr = "\n".join(
                ["{____classname}(", ",\n".join(args_kv), indent + ")"]
            )
        else:  # no arguments
            reprformatstr = "{____classname}()"

        # add classname to format args
        reprformatargs = init_args.copy()
        reprformatargs.update({"____classname": classname})

        reprstring = (reprformatstr).format(**reprformatargs)
        return reprstring
